import setuptools


def get_version():
    """get_version - thanks to
       https://milkr.io/kfei/5-common-patterns-to-version-your-Python-package
    """
    import os
    import re
    VERSIONFILE = os.path.join('Acquire', '__init__.py')
    initfile_lines = open(VERSIONFILE, 'rt').readlines()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in initfile_lines:
        mo = re.search(VSRE, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError('Unable to find version string in %s.' % (VERSIONFILE,))


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="acquire",
    version=get_version(),
    python_requires='>=3.6.0',
    author="Christopher Woods",
    author_email="chryswoods@gmail.com",
    description="A serverless identity, access, accounting, storage and "
                "compute management system",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/chryswoods/acquire",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License  ",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
           "acquire_login = Acquire.Client.Scripts.__acquire_login__:main"
        ]
    },
    install_requires=[
        "pyotp>=2",
        "cachetools>=3",
        "tblib>=1.2",
        "cryptography>=2",
        "pyyaml>=3.0",
        "requests>=2.10",
        "qrcode[pil]>=5.0"
    ]
)




__all__ = ["Request"]


class Request:
    """This is the base class for all resource request classses. These
       classes are used to transmit information about a resource
       request, together with the user authorisation and account
       from which payment for the resource should be taken
    """
    def __init__(self):
        """Construct the resource request"""
        pass

    def is_null(self):
        """Return whether or not this request is null"""
        return True

    def to_data(self):
        """Return this class as a json-serialisable dictionary"""
        data = {}

        data["class"] = str(self.__class__.__name__)

        return data

    @staticmethod
    def from_data(data):
        """Construct a Request from the data in the json-deserialised
           dictionary
        """

        if (data and len(data) > 0):
            try:
                classname = data["class"]
            except:
                return Request()

            if classname == "FileWriteRequest":
                from ._filewriterequest import FileWriteRequest \
                                            as _FileWriteRequest
                return _FileWriteRequest.from_data(data)
            elif classname == "RunRequest":
                from ._runrequest import RunRequest as _RunRequest
                return _RunRequest.from_data(data)
            else:
                raise TypeError("Unknown type '%s'" % classname)

        return Request()

    def _from_data(self, data):
        """Call this function from derived classes to load data
           into this Request
        """
        pass




from Acquire.Service import Service as _Service

__all__ = ["IdentityService"]


class IdentityService(_Service):
    """This is a specialisation of Service for Identity Services"""
    def __init__(self, other=None):
        if isinstance(other, _Service):
            from copy import copy as _copy
            self.__dict__ = _copy(other.__dict__)

            if not self.is_identity_service():
                from Acquire.Identity import IdentityServiceError
                raise IdentityServiceError(
                    "Cannot construct an IdentityService from "
                    "a service which is not an identity service!")
        else:
            _Service.__init__(self)

    def _call_local_function(self, function, args):
        """Internal function called to short-cut local 'remote'
           function calls
        """
        from identity.route import identity_functions as _identity_functions
        from admin.handler import create_handler as _create_handler
        handler = _create_handler(_identity_functions)
        return handler(function, args)




import uuid as _uuid
import json as _json
import datetime as _datetime

from cachetools import LRUCache as _LRUCache
from cachetools import cached as _cached

__all__ = ["Service"]

_cache_service_user = _LRUCache(maxsize=5)


class ServiceError(Exception):
    pass


def _create_service_user(service_type=None):
    """This function is called to create the service user account for
       this service. The service user is the actual user who manages
       and authorises everything for this service. It it not possible
       to login as this user from outside the service. Instead, you
       login as one of the admin accounts, and then instruct the
       service user to perform the various tasks. There is one, and
       only one service user account for each service. It is as
       unchanging as the service UID. If the service user account
       already exists, then this function will raise an exception.

       If successful, this will return the username, UID and login secrets
       of the new service account
    """
    from Acquire.Identity import UserAccount as _UserAccount
    from Acquire.Crypto import PrivateKey as _PrivateKey
    from Acquire.Crypto import OTP as _OTP
    from Acquire.ObjectStore import ObjectStore as _ObjectStore
    from Acquire.ObjectStore import Mutex as _Mutex
    from Acquire.Service import get_service_account_bucket as \
        _get_service_account_bucket
    from Acquire.Service import ServiceAccountError

    bucket = _get_service_account_bucket()

    if service_type is None:
        username = "service_principal"
    else:
        username = "%s_principal" % service_type

    service_account = _UserAccount(username)

    password = _PrivateKey.random_passphrase()

    # generate the encryption keys and otp secret
    privkey = _PrivateKey()
    pubkey = privkey.public_key()
    otp = _OTP()

    # save the encrypted private key (encrypted using the user's password)
    # and encrypted OTP secret (encrypted using the public key)
    service_account.set_keys(privkey.bytes(password), pubkey.bytes(),
                             otp.encrypt(pubkey))

    # can only do this once
    account_key = "identity/accounts/%s" % service_account.sanitised_name()
    mutex = _Mutex(account_key)

    try:
        service_user = _ObjectStore.get_object_from_json(bucket, account_key)
    except:
        service_user = None

    if service_user is not None:
        raise ServiceAccountError(
            "The Service User Account can only be created ONCE per service!")

    # save the new account details
    _ObjectStore.set_object_from_json(bucket, account_key,
                                      service_account.to_data())

    # need to update the "whois" database with the uuid of this user
    _ObjectStore.set_string_object(bucket,
                                   "identity/whois/%s" %
                                   service_account.uuid(),
                                   service_account.username())

    mutex.unlock()

    # everything is ok - add this admin user to the admin_users
    # dictionary
    user_secret = {"password": password,
                   "otpsecret": otp._secret}

    return (username, service_account.uid(), user_secret)


@_cached(_cache_service_user)
def _login_service_user(service_uid):
    """Login to the service user account for the service with
       UID 'service_uid'. The service user account is an
       account that provides full control for this service. The "admin"
       accounts invoke actions by logging into the service account and
       authorising actions using that account. It is not possible to
       login as the service user from outside the service. It is an
       account that is internal to the service.
    """
    from Acquire.Service import get_this_service as _get_this_service
    from Acquire.Client import User as _User
    from Acquire.Crypto import OTP as _OTP
    from Acquire.Identity import LoginSession as _LoginSession

    service = _get_this_service(need_private_access=True)

    if service.uid() != service_uid:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "You cannot login to the service account for '%s' from "
            "the service running at '%s'" % (service.uid(),
                                             service_uid))

    secrets = service.skeleton_key().decrypt(service.service_user_secrets())
    secrets = _json.loads(secrets)

    password = secrets["password"]
    otpsecret = secrets["otpsecret"]

    user = _User(user_uid=service.service_user_uid(),
                 identity_url=service.canonical_url())

    user.request_login(_is_local=True)
    short_uid = _LoginSession.to_short_uid(user.session_uid())

    login_args = {"short_uid": short_uid,
                  "username": user.username(),
                  "password": password,
                  "otpcode": _OTP(otpsecret).generate()}

    secrets = None

    from ._function import call_function as _call_function
    result = _call_function(service.canonical_url(), function="login",
                            args=login_args)

    login_args = None

    if result["status"] != 0:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "Error logging into the admin account: %s" % result["message"])

    user.wait_for_login()

    return user


class Service:
    """This class represents a service in the system. Services
       will either be identity services, access services,
       storage services or accounting services.
    """
    def __init__(self, service_type=None, service_url=None):
        """Construct a new service of the specified type, with
           the specified URL."""
        self._service_type = service_type
        self._service_url = service_url
        self._canonical_url = service_url
        self._skeleton_key = None
        self._uid = None
        self._pubcert = None
        self._pubkey = None
        self._service_user_name = None
        self._service_user_uid = None
        self._service_user_secrets = None

        if self._service_type:
            if self._service_type not in ["identity", "access", "compute",
                                          "accounting", "storage"]:
                raise ServiceError("Services of type '%s' are not allowed!" %
                                   self._service_type)

            from Acquire.Crypto import PrivateKey as _PrivateKey

            self._uid = str(_uuid.uuid4())
            self._skeleton_key = _PrivateKey()

            self._privkey = _PrivateKey()
            self._privcert = _PrivateKey()

            self._pubkey = self._privkey.public_key()
            self._pubcert = self._privcert.public_key()

            # generate 'dummy' old keys - these will be replaced as
            # the real keys are updated
            self._lastkey = self._privkey
            self._lastcert = self._privcert

            from Acquire.ObjectStore import get_datetime_now as \
                _get_datetime_now
            self._last_key_update = _get_datetime_now()
            self._key_update_interval = 3600 * 24 * 7  # update keys weekly

            (username, uid, secrets) = _create_service_user(self._service_type)

            self._service_user_name = username
            self._service_user_uid = uid
            self._service_user_secrets = self._skeleton_key.encrypt(
                                                    _json.dumps(secrets))

    def __str__(self):
        if self._uid is None:
            return "%s(NULL)" % self.__class__.__name__
        else:
            return "%s(url=%s, uid=%s)" % (self.__class__.__name__,
                                           self.canonical_url(),
                                           self.uid())

    def uuid(self):
        """Synonym for uid"""
        return self.uid()

    def uid(self):
        """Return the uuid of this service. This MUST NEVER change, as
           the UID uniquely identifies this service to all other
           services
        """
        if self._uid is None:
            self.refresh_keys()

        return self._uid

    def service_type(self):
        """Return the type of this service"""
        return self._service_type

    def bucket(self):
        """Return the bucket you can use to read/write data to the
           object store associated with this service account
        """
        from Acquire.Service import get_service_account_bucket as \
            _get_service_account_bucket
        return _get_service_account_bucket()

    def is_locked(self):
        """Return whether or not this service object is locked. Locked
           service objects don't contain copies of any private keys,
           and can be safely shared as a means of distributing public
           keys and certificates
        """
        return self._skeleton_key is None

    def is_unlocked(self):
        """Return whether or not this service object is unlocked. Unlocked
           service objects have access to the skeleton key and other private
           keys. They should only run on the service. Locked service objects
           are what are returned by services to provide public keys and
           public certificates
        """
        return self._skeleton_key is not None

    def get_trusted_service(self, service_url=None, service_uid=None):
        """Return the trusted service info for the service with specified
           service_url or service_uid"""
        from ._get_services import get_trusted_service as \
            _get_trusted_service

        return _get_trusted_service(service_url=service_url,
                                    service_uid=service_uid)

    def assert_unlocked(self):
        """Assert that this service object is unlocked"""
        if self.is_locked():
            raise ServiceError(
                "Cannot complete operation as the service account '%s' "
                "is locked" % str(self))

    def assert_admin_authorised(self, authorisation, resource=None):
        """Validate that the passed authorisation is valid for the
           (optionally) specified resource, and that this has been
           authorised by one of the admin accounts of this service
        """
        if authorisation.identity_uid() != self.uid():
            from Acquire.Identity import AuthorisationError
            raise AuthorisationError(
                "The authorisation has not been signed by one of the "
                "admin accounts on service '%s'" % str(self))

        from Acquire.Service import get_admin_users as _get_admin_users
        admin_users = _get_admin_users()

        if authorisation.user_uid() not in admin_users:
            raise AuthorisationError(
                "The authorisation has not been signed by one of the "
                "admin accounts on service '%s'" % str(self))

        authorisation.verify(resource)

    def last_key_update(self):
        """Return the datetime when the key and certificate of this
           service were last updated
        """
        try:
            return self._last_key_update
        except:
            return None

    def key_update_interval(self):
        """Return the time delta between server key updates"""
        try:
            return _datetime.timedelta(seconds=self._key_update_interval)
        except:
            return _datetime.timedelta(seconds=1)

    def should_refresh_keys(self):
        """Return whether the keys and certificates need to be refreshed
           - i.e. more than 'key_update_interval' has passed since the last
           key update
        """
        try:
            from Acquire.ObjectStore import get_datetime_now as \
                _get_datetime_now

            return _get_datetime_now() > (self._last_key_update +
                                          self.key_update_interval())
        except:
            return True

    def refresh_keys(self):
        """Refresh the keys and certificates"""
        if self._uid is None:
            return

        if self.is_unlocked():
            # actually regenerate keys for the service - first save
            # the old private key (so we can decrypt data decrypted using
            # the old public key) and the old public certificate, so we
            # can verify data signed using the old private certificate
            self._lastkey = self._privkey
            self._lastcert = self._privcert

            # now generate a new key and certificate
            from Acquire.Crypto import PrivateKey as _PrivateKey
            self._privkey = _PrivateKey()
            self._privcert = _PrivateKey()
            self._pubkey = self._privkey.public_key()
            self._pubcert = self._privcert.public_key()

            # update the refresh time
            from Acquire.ObjectStore import get_datetime_now as \
                _get_datetime_now

            self._last_key_update = _get_datetime_now()
        else:
            from Acquire.Crypto import get_private_key as _get_private_key
            from ._function import call_function as _call_function

            # if our keys are old then pull the new ones from the server
            if self._pubcert is None:
                # we are initialising from scratch - hope this is over https
                response = _call_function(
                    self._service_url,
                    response_key=_get_private_key("function"))
            else:
                # ask for an updated Service, ensuring the service responds
                # with a signature that we know was (once) valid
                response = _call_function(
                    self._service_url,
                    response_key=_get_private_key("function"),
                    public_cert=self._pubcert)

            service = Service.from_data(response["service_info"],
                                        verify_data=True)

            if service.uid() != self.uid():
                raise ServiceError(
                    "Cannot update the service as the UID has changed. We "
                    "cannot move from %s to %s. Contact an administrator." %
                    (str(self), str(service)))

            # everything should be ok. Update this object with the new
            # keys and data
            from copy import copy as _copy
            self.__dict__ = _copy(service.__dict__)

    def can_identify_users(self):
        """Return whether or not this service can identify users.
           Most services can, at a minimum, identify their admin
           users. However, only true Identity Services can register
           and manage normal users
        """
        return True

    def is_identity_service(self):
        """Return whether or not this is an identity service"""
        if self._service_type:
            return self._service_type == "identity"
        else:
            return False

    def is_access_service(self):
        """Return whether or not this is an access service"""
        if self._service_type:
            return self._service_type == "access"
        else:
            return False

    def is_accounting_service(self):
        """Return whether or not this is an accounting service"""
        if self._service_type:
            return self._service_type == "accounting"
        else:
            return False

    def is_compute_service(self):
        """Return whether or not this is a compute service"""
        if self._service_type:
            return self._service_type == "compute"
        else:
            return False

    def is_storage_service(self):
        """Return whether or not this is a storage service"""
        if self._service_type:
            return self._service_type == "storage"
        else:
            return False

    def service_url(self):
        """Return the URL used to access this service"""
        return self._service_url

    def canonical_url(self):
        """Return the canonical URL for this service (this is the URL the
           service thinks it has, and which it has used to register itself
           with all other services)
        """
        return self._canonical_url

    def hostname(self):
        """Return the hostname of the canonical URL that provides
           this service
        """
        if self._uid is None:
            return None

        from urllib.parse import urlparse as _urlparse
        return _urlparse(self.canonical_url()).hostname

    def uses_https(self):
        """Return whether or not the canonical URL of this service
           is connected to via https
        """
        if self._uid is None:
            return False

        from urllib.parse import urlparse as _urlparse
        return _urlparse(self.canonical_url()).scheme == "https"

    def update_service_url(self, service_url):
        """Update the service url to be 'service_url'"""
        self._service_url = str(service_url)

    def service_user_uid(self):
        """Return the UID of the service user account for this service"""
        return self._service_user_uid

    def service_user_name(self):
        """Return the name of the service user account for this service"""
        return self._service_user_name

    def service_user_secrets(self):
        """Return the (encrypted) secrets for the service user account.
           These will only be returned if you have unlocked this service.
           You need access to the skeleton key to decrypt these secrets
        """
        self.assert_unlocked()
        return self._service_user_secrets

    def login_service_user(self):
        """Return a logged in Acquire.Client.User for the service user.
           This can only be called inside the service, and when you
           have unlocked this service object
        """
        self.assert_unlocked()
        return _login_service_user(self.uid())

    def service_user_account_uid(self, accounting_service_url=None,
                                 accounting_service=None):
        """Return the UID of the financial account associated with
           this service on the passed accounting service
        """
        from Acquire.Service import get_service_user_account_uid as \
            _get_service_user_account_uid

        if accounting_service is None:
            if accounting_service_url is None:
                raise ValueError(
                    "You must supply either an accounting service or "
                    "the URL of a valid accounting service!")

            accounting_service = self.get_trusted_service(
                                        accounting_service_url)

        if not accounting_service.is_accounting_service():
            raise TypeError(
                "The service '%s' is not an accounting service!"
                % str(accounting_service))

        return _get_service_user_account_uid(
                    accounting_service_uid=accounting_service.uid())

    def skeleton_key(self):
        """Return the skeleton key used by this service. This is an
           unchanging key which is stored internally, should never be
           shared outside the service, and which is used to encrypt
           all data. Unlocking the service involves loading and
           decrypting this skeleton key
        """
        self.assert_unlocked()
        return self._skeleton_key

    def private_key(self):
        """Return the private key (if it has been unlocked)"""
        self.assert_unlocked()
        return self._privkey

    def private_certificate(self):
        """Return the private signing certificate (if it has been unlocked)"""
        self.assert_unlocked()
        return self._privcert

    def public_key(self):
        """Return the public key for this service"""
        return self._pubkey

    def public_certificate(self):
        """Return the public signing certificate for this service"""
        return self._pubcert

    def last_key(self):
        """Return the old private key for this service (if it has
           been unlocked). This was the key used before the last
           key update, and we store it in case we have to decrypt
           data that was recently encrypted using the old public key
        """
        self.assert_unlocked()
        return self._lastkey

    def last_certificate(self):
        """Return the old public certificate for this service. This was the
           certificate used before the last key update, and we store it
           in case we need to verify data signed using the old private
           certificate
        """
        return self._lastcert

    def call_function(self, function, args=None):
        """Call the function 'func' on this service, optionally passing
           in the arguments 'args'. This is a simple wrapper around
           Acquire.Service.call_function which automatically
           gets the correct URL, encrypts the arguments using the
           service's public key, and supplies a key to encrypt
           the response (and automatically then decrypts the
           response)
        """
        if self._uid is None:
            from Acquire.Service import RemoteFunctionCallError
            raise RemoteFunctionCallError(
                "You cannot call the function '%s' on a null service!" %
                function)

        from Acquire.Crypto import get_private_key as _get_private_key
        from ._function import call_function as _call_function

        if self.should_refresh_keys():
            self.refresh_keys()

        return _call_function(self.canonical_url(), function=function,
                              args=args,
                              args_key=self.public_key(),
                              public_cert=self.public_certificate(),
                              response_key=_get_private_key("function"))

    def sign(self, message):
        """Sign the specified message"""
        return self.private_certificate().sign(message)

    def verify(self, signature, message):
        """Verify that this service signed the message"""
        self.public_certificate().verify(signature, message)

    def encrypt(self, message):
        """Encrypt the passed message"""
        return self.public_key().encrypt(message)

    def decrypt(self, message):
        """Decrypt the passed message"""
        return self.private_key().decrypt(message)

    def sign_data(self, data):
        """Sign the passed data, ready for transport. Data should be
           a json-serialisable dictionary. This will return a new
           json-serialisable dictionary, which will contain the
           signature and json-serialised original data, e.g. as;

           data = {"service_uid" : "SERVICE_UID",
                   "fingerprint" : "KEY_FINGERPRINT",
                   "signed_data" : "JSON_ENCODED_DATA",
                   "signature" : "SIG OF JSON_ENCODED_DATA"}
        """
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string

        data = _json.dumps(data)

        return {"service_uid": str(self.uid()),
                "canonical_url": str(self.canonical_url()),
                "fingerprint": str(self.private_certificate().fingerprint()),
                "signed_data": data,
                "signature": _bytes_to_string(self.sign(data))
                }

    def verify_data(self, data):
        """Verify the passed data has been signed by this service. The
           passed data should have the same format as that produced
           by 'sign_data'. If the data is verified then this will
           return a json-deserialised dictionary of the verified data.
           Note that the 'service_uid' should match the UID of this
           service. The data should also contain the fingerprint of the
           key used to encrypt the data, enabling the service to
           perform key rotation and management.
        """
        try:
            from Acquire.ObjectStore import string_to_bytes as _string_to_bytes

            service_uid = data["service_uid"]
            fingerprint = data["fingerprint"]
            signature = _string_to_bytes(data["signature"])
            data = data["signed_data"]
        except Exception as e:
            raise ServiceError(
                "The signed data is not of the correct format: %s" % str(e))

        if service_uid != self.uid():
            raise ServiceError(
                "Cannot verify the data as it wasn't signed for this "
                "service - unmatched service UID: %s versus %s" %
                (service_uid, self.uid()))

        if fingerprint != self.public_certificate().fingerprint():
            raise ServiceError(
                "Cannot verify the data as we don't recognise the "
                "fingerprint of the signing key: %s versus %s" %
                (fingerprint, self.public_certificate().fingerprint()))

        self.verify(signature, data)
        return _json.loads(data)

    def encrypt_data(self, data):
        """Encrypt the passed data, ready for transport to the service.
           Data should be a json-serialisable dictionary. This will
           return a new json-serialisable dictionary, which will contain
           the UID of the service this should be sent to (together with
           the canonical URL, which enables this data to be forwarded
           to where it needs to go), and the encrypted
           data, e.g. as;

           data = {"service_uid" : "SERVICE_UID",
                   "canonical_url" : "CANONICAL_URL",
                   "fingerprint" : "KEY_FINGERPRINT",
                   "encrypted_data" : "ENCRYPTED_DATA"}
        """
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
        return {"service_uid": str(self.uid()),
                "canonical_url": str(self.canonical_url()),
                "fingerprint": str(self.public_key().fingerprint()),
                "encrypted_data": _bytes_to_string(
                                    self.encrypt(_json.dumps(data)))
                }

    def decrypt_data(self, data):
        """Decrypt the passed data that has been encrypted and sent to
           this service (encrypted via the 'encrypt_data' function).
           This will return a json-deserialisable dictionary. Note that
           the 'service_uid' should match the UID of this
           service. The data should also contain the fingerprint of the
           key used to encrypt the data, enabling the service to
           perform key rotation and management.
        """
        if isinstance(data, str):
            data = _json.loads(data)

        try:
            from Acquire.ObjectStore import string_to_bytes as _string_to_bytes
            service_uid = data["service_uid"]
            fingerprint = data["fingerprint"]
            data = _string_to_bytes(data["encrypted_data"])
        except Exception as e:
            raise ServiceError(
                "The encrypted data is not of the correct format: %s" % str(e))

        if service_uid != self.uid():
            raise ServiceError(
                "Cannot decrypt the data as it wasn't encrypted for this "
                "service - unmatched service UID: %s versus %s" %
                (service_uid, self.uid()))

        if fingerprint != self.private_key().fingerprint():
            raise ServiceError(
                "Cannot decrypt the data as we don't recognise the "
                "fingerprint of the encryption key: %s versus %s" %
                (fingerprint, self.private_key().fingerprint()))

        data = self.decrypt(data)
        return _json.loads(data)

    def dump_keys(self, include_old_keys=False):
        """Return a dump of the current key and certificate, so that
           we can keep a record of all keys that have been used. The
           returned json-serialisable dictionary contains the keys,
           their fingerprints, and the datetime when they were
           generated. If this is run on the service, then the keys
           are encrypted the password which is encrypted using the
           master key
        """
        dump = {}

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        dump["datetime"] = _datetime_to_string(self._last_key_update)

        if self.is_unlocked():
            from Acquire.Crypto import PrivateKey as _PrivateKey
            from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
            ranpas = _PrivateKey.random_passphrase()

            key = self.private_key()
            cert = self.private_certificate()

            dump[key.fingerprint()] = key.to_data(ranpas)
            dump[cert.fingerprint()] = cert.to_data(ranpas)

            if include_old_keys:
                key = self.last_key()
                cert = self.last_certificate()
                dump[key.fingerprint()] = key.to_data(ranpas)
                dump[cert.fingerprint()] = cert.to_data(ranpas)

            ranpas = _bytes_to_string(self._skeleton_key.encrypt(ranpas))
            dump["encrypted_passphrase"] = ranpas
        else:
            key = self.public_key()
            cert = self.public_certificate()

            dump[key.fingerprint()] = key.to_data()
            dump[cert.fingerprint()] = cert.to_data()

            if include_old_keys:
                cert = self.last_certificate()
                dump[cert.fingerprint()] = cert.to_data()

        return dump

    def load_keys(self, data):
        """Return the keys that were dumped by 'self.dump_keys()'.
           This returns a dictionary of the keys and datetime that
           they were created, indexed by their key fingerprints
        """
        # get all of the key fingerprints in this dictionary
        fingerprints = []
        for key in data.keys():
            if key not in ["datetime", "encrypted_passphrase"]:
                fingerprints.append(key)

        # now unpack everything
        result = {}

        from Acquire.ObjectStore import string_to_datetime as \
            _string_to_datetime
        result["datetime"] = _string_to_datetime(data["datetime"])

        if self.is_unlocked():
            from Acquire.Crypto import PrivateKey as _PrivateKey
            from Acquire.ObjectStore import string_to_bytes as _string_to_bytes
            ranpas = self._skeleton_key.decrypt(
                            _string_to_bytes(data["encrypted_passphrase"]))

            for fingerprint in fingerprints:
                result[fingerprint] = _PrivateKey.from_data(data[fingerprint],
                                                            ranpas)
        else:
            from Acquire.Crypto import PublicKey as _PublicKey
            for fingerprint in fingerprints:
                result[fingerprint] = _PublicKey.from_data(data[fingerprint])

        return result

    def whois(self, username=None, user_uid=None, session_uid=None):
        """Do a whois lookup to map from username to user_uid or
           vice versa. If 'session_uid' is provided, then also validate
           that this is a correct login session, and return also
           the public key and signing certificate for this login session.

           This should return a dictionary with the following keys
           optionally contained;

           username = name of the user
           user_uid = uid of the user
           public_key = public key for the session with uid 'session_uid'
           public_cert = public certificate for that login session
        """

        if (username is None) and (user_uid is None):
            from Acquire.Identity import IdentityServiceError
            raise IdentityServiceError(
                    "You must supply either a username "
                    "or a user's UID for a lookup")

        response = None

        if session_uid is None:
            args = {}
        else:
            args = {"session_uid": str(session_uid)}

        try:
            from Acquire.Crypto import get_private_key as _get_private_key
            from ._function import call_function as _call_function

            if username:
                args["username"] = str(username)
                response = _call_function(
                                self.service_url(), "whois",
                                public_cert=self.public_certificate(),
                                response_key=_get_private_key("function"),
                                args=args)
                lookup_uid = response["user_uid"]
            else:
                lookup_uid = None

            if user_uid:
                args["user_uid"] = str(user_uid)
                response = _call_function(
                    self.service_url(), "whois",
                    public_cert=self.public_certificate(),
                    response_key=_get_private_key("function"),
                    args=args)
                lookup_username = response["username"]
            else:
                lookup_username = None

        except Exception as e:
            from Acquire.Identity import IdentityServiceError
            raise IdentityServiceError("Failed whois lookup: %s" % str(e))

        if username is None:
            username = lookup_username

        elif (lookup_username is not None) and (username != lookup_username):
            from Acquire.Identity import IdentityServiceError
            raise IdentityServiceError(
                "Disagreement of the user who matches "
                "UID=%s. We think '%s', but the identity service says '%s'" %
                (user_uid, username, lookup_username))

        if user_uid is None:
            user_uid = lookup_uid

        elif (lookup_uid is not None) and (user_uid != lookup_uid):
            from Acquire.Identity import IdentityServiceError
            raise IdentityServiceError(
                    "Disagreement of the user's UID for user "
                    "'%s'. We think %s, but the identity service says %s" %
                    (username, user_uid, lookup_uid))

        result = response

        try:
            result["public_key"] = _PublicKey.from_data(
                                            response["public_key"])
        except:
            pass

        try:
            result["public_cert"] = _PublicKey.from_data(
                                            response["public_cert"])
        except:
            pass

        return result

    def is_evolution_of(self, other):
        """Return whether or not this service is an evolution of 'other'.
           Evolving means that this service is the same service as 'other',
           but at a later point in time with newer keys
        """
        if self.validation_string() == other.validation_string():
            return True
        elif self.canonical_url() != other.canonical_url():
            return False
        elif self.uid() != other.uid():
            return False

        print("WARNING - NEED TO IMPLEMENT KEY EVOLUTION COMPARISON")

        return False

    def validation_string(self):
        """Return a string created from this object that can be signed
           to verify that all information was transmitted correctly
        """
        return "%s:%s:%s:%s:%s:%s:%s:%s:%s" % (
            self._uid, self.canonical_url(), self._service_type,
            self._pubcert.fingerprint(), self._pubkey.fingerprint(),
            self._lastcert.fingerprint(), self._service_user_uid,
            self._last_key_update.isoformat(), self._key_update_interval)

    def to_data(self, password=None):
        """Serialise this key to a dictionary, using the supplied
           password to encrypt the private key and certificate"""

        data = {}

        data["uid"] = self._uid
        data["service_type"] = self._service_type
        data["service_url"] = self._service_url

        data["public_certificate"] = self._pubcert.to_data()
        data["public_key"] = self._pubkey.to_data()

        from Acquire.Crypto import PublicKey as _PublicKey
        from Acquire.ObjectStore import datetime_to_string as \
            _datetime_to_string

        if isinstance(self._lastcert, _PublicKey):
            data["last_certificate"] = self._lastcert.to_data()
        else:
            data["last_certificate"] = self._lastcert.public_key().to_data()

        data["last_key_update"] = _datetime_to_string(self._last_key_update)
        data["key_update_interval"] = self._key_update_interval

        data["service_user_name"] = self._service_user_name
        data["service_user_uid"] = self._service_user_uid

        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string

        if (self.is_unlocked()) and (password is not None):
            # only serialise private data if a password was provided
            from Acquire.Crypto import PrivateKey as _PrivateKey

            secret_passphrase = _PrivateKey.random_passphrase()
            key_data = self._privkey.to_data(secret_passphrase)
            lastkey_data = self._lastkey.to_data(secret_passphrase)
            cert_data = self._privcert.to_data(secret_passphrase)

            secret_data = {"passphrase": secret_passphrase,
                           "private_key": key_data,
                           "last_key": lastkey_data,
                           "private_certificate": cert_data}

            secret_data = self._skeleton_key.encrypt(_json.dumps(secret_data))

            data["secret_data"] = _bytes_to_string(secret_data)
            data["skeleton_key"] = self._skeleton_key.to_data(password)

            # the service user secrets are already encrypted
            data["service_user_secrets"] = _bytes_to_string(
                                            self._service_user_secrets)
        elif self.is_unlocked():
            # sign a validation string so that people can
            # check it has not been tampered with in transit
            v = self.validation_string()
            data["validation_string"] = v
            data["public_signature"] = _bytes_to_string(
                self._privcert.sign(message=v))
            data["last_signature"] = _bytes_to_string(
                self._lastcert.sign(message=v))

        return data

    @staticmethod
    def from_data(data, password=None, verify_data=False):
        """Deserialise this object from the passed data. This will
           only deserialise the private key and private certificate
           if the password is supplied.

           If 'verify_data' is True, then extract the signature of the
           data and verify that that signature is correct. You should
           always verify data that has been transmitted over a network.
        """
        service = Service()

        from Acquire.Crypto import PublicKey as _PublicKey
        from Acquire.ObjectStore import string_to_datetime as \
            _string_to_datetime

        if password:
            from Acquire.Crypto import PrivateKey as _PrivateKey
            from Acquire.ObjectStore import string_to_bytes as _string_to_bytes

            # get the private info...
            service._service_user_secrets = _string_to_bytes(
                                                data["service_user_secrets"])

            service._skeleton_key = _PrivateKey.from_data(data["skeleton_key"],
                                                          password)

            secret = service._skeleton_key.decrypt(_string_to_bytes(
                                                   data["secret_data"]))

            secret = _json.loads(secret)

            passphrase = secret["passphrase"]
            service._privkey = _PrivateKey.from_data(secret["private_key"],
                                                     passphrase)
            service._lastkey = _PrivateKey.from_data(secret["last_key"],
                                                     passphrase)
            service._privcert = _PrivateKey.from_data(
                                            secret["private_certificate"],
                                            passphrase)

            try:
                service._lastcert = _PrivateKey.from_data(
                                                secret["last_certificate"],
                                                passphrase)
            except:
                service._lastcert = service._privcert
        else:
            service._skeleton_key = None
            service._privkey = None
            service._privcert = None
            service._lastkey = None
            service._lastcert = _PublicKey.from_data(data["last_certificate"])
            service._service_user_secrets = None

        service._uid = data["uid"]
        service._service_type = data["service_type"]
        service._service_url = data["service_url"]
        service._canonical_url = service._service_url

        service._service_user_uid = data["service_user_uid"]
        service._service_user_name = data["service_user_name"]

        service._pubkey = _PublicKey.from_data(data["public_key"])
        service._pubcert = _PublicKey.from_data(data["public_certificate"])

        service._last_key_update = _string_to_datetime(data["last_key_update"])
        service._key_update_interval = float(data["key_update_interval"])

        if service.is_identity_service():
            from Acquire.Identity import IdentityService as _IdentityService
            service = _IdentityService(service)
        elif service.is_access_service():
            from Acquire.Access import AccessService as _AccessService
            service = _AccessService(service)
        elif service.is_compute_service():
            from Acquire.Compute import ComputeService as _ComputeService
            service = _ComputeService(service)
        elif service.is_storage_service():
            from Acquire.Storage import StorageService as _StorageService
            service = _StorageService(service)
        elif service.is_accounting_service():
            from Acquire.Accounting import AccountingService \
                                        as _AccountingService
            service = _AccountingService(service)

        if verify_data:
            # the service was transmitted with a signature from both
            # certificates - make sure that the signature was correct.
            # This stops someone changing the certificates, keys or
            # any other data about the service while in transit
            try:
                from Acquire.ObjectStore import string_to_bytes \
                    as _string_to_bytes

                vr = data["validation_string"]
                v = service.validation_string()
                assert(v == vr)

                sig = _string_to_bytes(data["public_signature"])
                service._pubcert.verify(signature=sig, message=v)
                sig = _string_to_bytes(data["last_signature"])
                service._lastcert.verify(signature=sig, message=v)
            except Exception as e:
                from Acquire.Crypto import SignatureVerificationError
                raise SignatureVerificationError(
                    "Cannot verify that the returned data for service '%s' "
                    "has not been tampered with - the signature is "
                    "incorrect: Error = %s" %
                    (service.canonical_url(), str(e)))

            service._verified_data = True

        return service




import uuid as _uuid
import datetime as _datetime
from copy import copy as _copy
from enum import Enum as _Enum

from ._errors import TransactionError, UnbalancedLedgerError, \
                     UnmatchedReceiptError, UnmatchedRefundError, \
                     LedgerError

__all__ = ["TransactionRecord", "TransactionState"]


class TransactionState(_Enum):
    """This class holds an enum of the current state of a transaction"""
    DIRECT = "DR"           # direct transaction, no receipts etc.
    PROVISIONAL = "PR"      # provisional transaction, needs receipt
    RECEIPTING = "RC"       # in process of being receipted...
    RECEIPTED = "RD"        # has been receipted
    REFUNDING = "RF"        # in process of being refunded...
    REFUNDED = "RR"         # has been refunded


class TransactionRecord:
    """This class holds a record of a transaction that has already been
       written to the accounting ledger. This records a unique ID, the
       datetime of the entry, the value, the two accounts involved in the
       transaction (debit account to credit account), a description of what
       the transaction refers to, and who/how the transaction was authorised.
       If 'is_provisional' then this is a provisional transaction that is
       recorded as a liability for the debtor and a future income for the
       creditor. This is confirmed by creating a receipt via "receipt_for"
       by passing the UID for the transaction this receipts, and the actual
       'value' of the receipt. Note that the actual value CANNOT exceed the
       original provisional value that was agreed by the debtor
    """
    def __init__(self, uid=None, bucket=None):
        """Load the transaction record from the object store using the
           passed UID
        """
        if uid:
            self._load_transaction(uid, bucket)
        else:
            self._debit_note = None
            self._credit_note = None
            self._transaction_state = None
            self._refund = None
            self._receipt = None

    def __str__(self):
        """Return a string representation of this transaction"""
        if self.is_null():
            return "TransactionRecord::null"

        return "[%s]: transferred %s from %s to %s | %s" % \
               (self.description(),
                self.value(), self.debit_note().account_uid(),
                self.credit_note().account_uid(),
                self._transaction_state.value)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._debit_note == other._debit_note and \
                   self._credit_note == other._credit_note and \
                   self._transaction_state == other._transaction_state
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def reload(self):
        """Reload this transaction record from the object store. This is
           necessary if, e.g., the state of the record has been updated
        """
        self._load_transaction(self.uid())

    def is_null(self):
        """Return whether or not this is a null record"""
        return self._debit_note is None

    def description(self):
        """Return the description of this transaction"""
        if self.is_null():
            return None
        else:
            return self.transaction().description()

    def value(self):
        """Return the value of this transaction"""
        if self.is_null():
            return 0
        else:
            return self.transaction().value()

    def uid(self):
        """Return the UID for this transaction record"""
        if self.is_null():
            return None
        else:
            return self.debit_note().uid()

    def transaction(self):
        """Return the transaction underlying this record"""
        if self.is_null():
            return None
        else:
            return self.debit_note().transaction()

    def transaction_state(self):
        """Return the current state of this transaction"""
        return self._transaction_state

    def assert_matching_refund(self, refund):
        """Assert that the passed refund matches this transaction"""
        if self.is_null():
            if not refund.is_null():
                raise UnmatchedRefundError(
                        "%s does not match a null TransactionRecord!" %
                        str(refund))

        match = True
        errors = []

        if refund.transaction_uid() != self.uid():
            match = False
            errors.append("UID: %s != %s" % (refund.transaction_uid(),
                                             self.uid()))
        elif refund.debit_account_uid() != self.debit_account_uid():
            match = False
            errors.append("DEBIT: %s != %s" % (refund.debit_account_uid(),
                                               self.debit_account_uid()))
        elif refund.credit_account_uid() != self.credit_account_uid():
            match = False
            errors.append("CREDIT: %s != %s" % (refund.credit_account_uid(),
                                                self.credit_account_uid()))
        elif refund.value() != self.value():
            match = False
            errors.append("VALUE: %s != %s" % (refund.value(), self.value()))

        if not match:
            raise UnmatchedRefundError(
                "The refund '%s' does not match the transaction '%s': %s" %
                (str(refund), str(self), " | ".join(errors)))

    def assert_matching_receipt(self, receipt):
        """Assert that the passed receipt matches this transaction"""
        if self.is_null():
            if not receipt.is_null():
                raise UnmatchedReceiptError(
                        "%s does not match a null TransactionRecord!" %
                        str(receipt))

        match = True
        errors = []

        if receipt.transaction_uid() != self.uid():
            match = False
            errors.append("UID: %s != %s" % (receipt.transaction_uid(),
                                             self.uid()))
        elif receipt.debit_account_uid() != self.debit_account_uid():
            match = False
            errors.append("DEBIT: %s != %s" % (receipt.debit_account_uid(),
                                               self.debit_account_uid()))
        elif receipt.credit_account_uid() != self.credit_account_uid():
            match = False
            errors.append("CREDIT: %s != %s" % (receipt.credit_account_uid(),
                                                self.credit_account_uid()))
        elif receipt.value() != self.value():
            match = False
            errors.append("VALUE: %s != %s" % (receipt.value(), self.value()))

        if not match:
            raise UnmatchedReceiptError(
                "The receipt '%s' does not match the transaction '%s': %s" %
                (str(receipt), str(self), " | ".join(errors)))

    def credit_account_uid(self):
        """Return the UID of the account to which value has been credited"""
        if self.is_null():
            return None
        else:
            return self.credit_note().account_uid()

    def credit_note(self):
        """Return the credit note for this transaction. This is the note
           recording that value has been credited to an account. A
           TransactionRecord is the pairing of a DebitNote with a CreditNote
        """
        return self._credit_note

    def debit_note(self):
        """Return the debit note for this transaction. This is the note
           recording that value has been debited to an account. A
           TransactionRecord is the pairing of a DebitNote with a CreditNote
        """
        return self._debit_note

    def debit_account_uid(self):
        """Return the UID of the account from which value has been debited"""
        if self.is_null():
            return None
        else:
            return self.debit_note().account_uid()

    def datetime(self):
        """Return the datetime when this transaction was applied"""
        if self.is_null():
            return None
        else:
            return self.debit_note().datetime()

    def is_direct(self):
        """Return whether or not this transaction was direct (so was not
           provisional and so didn't need a receipt)
        """
        return self._transaction_state == TransactionState.DIRECT

    def is_receipted(self):
        """Return whether or not this transaction has been receipted"""
        return self._transaction_state == TransactionState.RECEIPTED

    def is_refunded(self):
        """Return whether or not this transaction has been refunded"""
        return self._transaction_state == TransactionState.REFUNDED

    def is_provisional(self):
        """Return whether or not this transaction is provisional"""
        return self._transaction_state == TransactionState.PROVISIONAL

    def is_refund(self):
        """Return whether or not this transaction is a refund"""
        return self._refund is not None

    def is_receipt(self):
        """Return whether or not this transaction is a receipt"""
        return self._receipt is not None

    def get_refund_info(self):
        """Return the reason for the refund"""
        return self._refund

    def get_receipt_info(self):
        """Return the receipt underlying this transaction"""
        return self._receipt

    def original_transaction_record(self):
        """If this is a receipt or refund transaction then return the
           original transaction record that this is receipting or refunding.
           Otherwise returns a null TransactionRecord
        """
        if self.is_receipt():
            return TransactionRecord(self.get_receipt_info().transaction_uid())
        elif self.is_refund():
            return TransactionRecord(self.get_refund_info().transaction_uid())
        else:
            return TransactionRecord()

    def original_transaction(self):
        """If this is a receipt or refund transaction then return the
           original transaction that this is receipting or refunding.
           Otherwise returns a null Transaction
        """
        if self.is_receipt() or self.is_refund():
            return self.original_transaction_record().transaction()
        else:
            from Acquire.Accounting import Transaction as _Transaction
            return _Transaction()

    def _load_transaction(self, uid, bucket=None):
        """Load this transaction from the object store"""
        from Acquire.Accounting import Ledger as _Ledger
        self.__dict__ = _copy(_Ledger.load_transaction(
                                        uid, bucket=bucket).__dict__)

    def _save_transaction(self, bucket=None):
        """Save this transaction to the object store"""
        from Acquire.Accounting import Ledger as _Ledger
        _Ledger.save_transaction(self, bucket=bucket)

    @staticmethod
    def load_test_and_set(uid, expected_state, new_state,
                          bucket=None):
        """Static method to load up the Transaction record associated with
           the passed UID, check that the transaction state matches
           'expected_state', and if it does, to update the transaction
           state to 'new_state'. This returns the loaded (and updated)
           transaction
        """
        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.Accounting import Ledger as _Ledger
        from Acquire.ObjectStore import Mutex as _Mutex

        try:
            mutex = _Mutex(uid, timeout=600, lease_time=600)
        except Exception as e:
            raise LedgerError("Cannot secure a Ledger mutex for transaction "
                              "'%s'. Error = %s" % (uid, str(e)))

        try:
            transaction = _Ledger.load_transaction(uid, bucket)

            if transaction.transaction_state() != expected_state:
                raise TransactionError(
                    "Cannot update the state of the transaction %s from "
                    "%s to %s as it is not in the expected state" %
                    (str(transaction), expected_state.value, new_state.value))

            transaction._transaction_state = new_state
        except:
            mutex.unlock()
            raise

        # now need to write anything back if the state isn't changed
        if expected_state == new_state:
            return transaction

        # make sure we have enough time remaining on the lease to be
        # able to write this result back to the object store...
        if mutex.seconds_remaining_on_lease() < 100:
            try:
                mutex.fully_unlock()
            except:
                pass

            return TransactionRecord.load_test_and_set(uid, expected_state,
                                                       new_state, bucket)

        try:
            _Ledger.save_transaction(transaction, bucket)
        except:
            mutex.unlock()
            raise

        return transaction

    @staticmethod
    def from_data(data):
        """Construct and return a new Transaction from the passed json-decoded
            dictionary
        """
        record = TransactionRecord()

        if (data and len(data) > 0):
            from Acquire.Accounting import CreditNote as _CreditNote
            from Acquire.Accounting import DebitNote as _DebitNote

            record._credit_note = _CreditNote.from_data(data["credit_note"])
            record._debit_note = _DebitNote.from_data(data["debit_note"])
            record._transaction_state = TransactionState(
                                            data["transaction_state"])

            if "refund" in data:
                from Acquire.Accounting import Refund as _Refund
                record._refund = _Refund.from_data(data["refund"])
            else:
                record._refund = None

            if "receipt" in data:
                from Acquire.Accounting import Receipt as _Receipt
                record._receipt = _Receipt.from_data(data["receipt"])
            else:
                record._receipt = None

        return record

    def to_data(self):
        """Return this transaction as a dictionary that can be
           encoded to json
        """
        data = {}

        if not self.is_null():
            data["credit_note"] = self._credit_note.to_data()
            data["debit_note"] = self._debit_note.to_data()
            data["transaction_state"] = self._transaction_state.value

            if self._refund is not None:
                data["refund"] = self._refund.to_data()

            if self._receipt is not None:
                data["receipt"] = self._receipt.to_data()

        return data




import os as _os

from enum import Enum as _Enum

from datetime import datetime as _datetime
import time as _time

# If we can, import socket to get the hostname and IP address
try:
    import socket as _socket
    _has_socket = True
except:
    _has_socket = False

__all__ = ["User", "username_to_uid",
           "uid_to_username", "get_session_keys"]


class _LoginStatus(_Enum):
    EMPTY = 0
    LOGGING_IN = 1
    LOGGED_IN = 2
    LOGGED_OUT = 3
    ERROR = 4


def _get_identity_url():
    """Function to discover and return the default identity url"""
    return "http://fn.acquire-aaai.com:8080/t/identity"


def _get_identity_service(identity_url=None):
    """Function to return the identity service for the system"""
    if identity_url is None:
        identity_url = _get_identity_url()

    from Acquire.Client import LoginError

    try:
        from Acquire.Client import Wallet as _Wallet
        service = _Wallet.get_service(service_url=identity_url)
    except Exception as e:
        from Acquire.Service import exception_to_string
        raise LoginError("Have not received the identity service info from "
                         "the identity service at '%s'\n\nCAUSE: %s" %
                         (identity_url, exception_to_string(e)))

    if not service.can_identify_users():
        raise LoginError(
            "You can only use a valid identity service to log in! "
            "The service at '%s' is a '%s'" %
            (identity_url, service.service_type()))

    if identity_url != service.service_url():
        service.update_service_url(identity_url)

    return service


def uid_to_username(user_uid, identity_url=None):
    """Function to return the username for the passed uid"""
    from Acquire.Client import Service as _Service
    service = _get_identity_service(identity_url=identity_url)

    response = service.call_function(
                function="whois", args={"user_uid": str(user_uid)})

    return response["username"]


def username_to_uid(username, identity_url=None):
    """Function to return the uid for the passed username"""
    from Acquire.Client import Service as _Service
    service = _get_identity_service(identity_url=identity_url)

    response = service.call_function(
                function="whois", args={"username": str(username)})

    return response["user_uid"]


def get_session_keys(username=None, user_uid=None, session_uid=None,
                     identity_url=None):
    """Function to return the session keys for the specified user"""
    if username is None and user_uid is None:
        raise ValueError("You must supply either the username or user_uid!")

    if session_uid is None:
        raise ValueError("You must supply a valid UID for a login session")

    service = _get_identity_service(identity_url=identity_url)

    args = {"username": username,
            "user_uid": user_uid,
            "session_uid": session_uid}

    response = service.call_function(function="whois", args=args)

    from Acquire.Client import PublicKey as _PublicKey

    try:
        response["public_key"] = _PublicKey.from_data(response["public_key"])
    except:
        pass

    try:
        response["public_cert"] = _PublicKey.from_data(response["public_cert"])
    except:
        pass

    return response


class User:
    """This class holds all functionality that would be used
       by a user to authenticate with and access the service.
       This represents a single client login, and is the
       user-facing part of Acquire
    """
    def __init__(self, username=None, user_uid=None,
                 identity_url=None, identity_uid=None):
        """Construct a null user"""
        self._username = username
        self._status = _LoginStatus.EMPTY
        self._identity_service = None

        if identity_url:
            self._identity_url = identity_url

        if identity_uid:
            self._identity_uid = identity_uid
        else:
            self._identity_uid = None

        if user_uid is not None:
            username = uid_to_username(user_uid, identity_url)

            if self._username is not None:
                if username != self._username:
                    from Acquire.Client import LoginError
                    raise LoginError(
                        "Disagreement of username for account with UID '%s'. "
                        "%s versus %s" % (user_uid, username, self._username))

            self._username = username

        self._user_uid = None

    def __str__(self):
        return "User(name='%s', status=%s)" % (self.username(), self.status())

    def __enter__(self):
        """Enter function used by 'with' statements'"""
        pass

    def __exit__(self, exception_type, exception_value, traceback):
        """Ensure that we logout"""
        self.logout()

    def __del__(self):
        """Make sure that we log out before deleting this object"""
        self.logout()

    def _set_status(self, status):
        """Internal function used to set the status from the
           string obtained from the LoginSession
        """
        if status == "approved":
            self._status = _LoginStatus.LOGGED_IN
        elif status == "denied":
            self._set_error_state("Permission to log in was denied!")
        elif status == "logged_out":
            self._status = _LoginStatus.LOGGED_OUT

    def username(self):
        """Return the username of the user"""
        return self._username

    def uid(self):
        """Return the UID of this user. This uniquely identifies the
           user across all systems
        """
        if self._user_uid is None:
            self._user_uid = username_to_uid(self.username(),
                                             self.identity_service_url())
        return self._user_uid

    def guid(self):
        """Return the global UID of the user. While the UID is highly
           likely to be unique, the GUID should be globally guaranteed
           to be unique. This is ensured by combining the UID of the
           user with the UID of the identity service that
           primarily identifies the user
        """
        return "%s@%s" % (self.identity_service().uid(), self.uid())

    def status(self):
        """Return the current status of this user"""
        return self._status

    def _check_for_error(self):
        """Call to ensure that this object is not in an error
           state. If it is in an error state then raise an
           exception"""
        if self._status == _LoginStatus.ERROR:
            from Acquire.Client import LoginError
            raise LoginError(self._error_string)

    def _set_error_state(self, message):
        """Put this object into an error state, displaying the
           passed message if anyone tries to use this object"""
        self._status = _LoginStatus.ERROR
        self._error_string = message

    def session_key(self):
        """Return the session key for the current login session"""
        self._check_for_error()

        try:
            return self._session_key
        except:
            return None

    def signing_key(self):
        """Return the signing key used for the current login session"""
        self._check_for_error()

        try:
            return self._signing_key
        except:
            return None

    def identity_service(self):
        """Return the identity service info object for the identity
           service used to validate the identity of this user
        """
        if self._identity_service:
            return self._identity_service

        identity_service = _get_identity_service(
                                self.identity_service_url())

        # if the user supplied the UID then validate this is correct
        # pylint: disable=assignment-from-none
        if self._identity_uid:
            if identity_service.uid() != self._identity_uid:
                from Acquire.Client import LoginError
                raise LoginError(
                    "The UID of the identity service at '%s', which is "
                    "%s, does not match that supplied by the user, '%s'. "
                    "You should double-check that the UID is correct, or "
                    "that you have supplied the correct identity_url" %
                    (self.identity_service_url(), identity_service.uid(),
                     self._identity_uid))
        else:
            self._identity_uid = identity_service.uid()
        # pylint: enable=assignment-from-none

        self._identity_service = identity_service

        return self._identity_service

    def identity_service_uid(self):
        """Return the UID of the identity service. The combination
           of user_uid+service_uid should uniquely identify this user
           account anywhere in the world
        """
        if self._identity_uid is not None:
            return self._identity_uid
        else:
            return self._identity_service.uid()

    def identity_service_url(self):
        """Return the URL to the identity service. This is the full URL
           to the service, minus the actual function to be called, e.g.
           https://function_service.com/t/identity
        """
        self._check_for_error()

        try:
            return self._identity_url
        except:
            # return the default URL - this should be discovered...
            return _get_identity_url()

    def login_url(self):
        """Return the URL that the user must connect to to authenticate
           this login session"""
        self._check_for_error()

        try:
            return self._login_url
        except:
            return None

    def login_qr_code(self):
        """Return a QR code of the login URL that the user must connect to
           to authenticate this login session"""
        self._check_for_error()

        try:
            return self._login_qrcode
        except:
            return None

    def session_uid(self):
        """Return the UID of the current login session. Returns None
           if there is no valid login session"""
        self._check_for_error()

        try:
            return self._session_uid
        except:
            return None

    def is_empty(self):
        """Return whether or not this is an empty login (so has not
           been used for anything yet..."""
        return self._status == _LoginStatus.EMPTY

    def is_logged_in(self):
        """Return whether or not the user has successfully logged in"""
        return self._status == _LoginStatus.LOGGED_IN

    def is_logging_in(self):
        """Return whether or not the user is in the process of loggin in"""
        return self._status == _LoginStatus.LOGGING_IN

    def logout(self):
        """Log out from the current session"""
        if self.is_logged_in() or self.is_logging_in():
            service = self.identity_service()

            # create a permission message that can be signed
            # and then validated by the user
            permission = "Log out request for %s" % self._session_uid
            signature = self.signing_key().sign(permission)

            from Acquire.ObjectStore import bytes_to_string \
                as _bytes_to_string

            args = {"username": self._username,
                    "session_uid": self._session_uid,
                    "permission": permission,
                    "signature": _bytes_to_string(signature)}

            result = service.call_function(function="logout", args=args)

            self._status = _LoginStatus.LOGGED_OUT

            return result

    def register(self, password, identity_url=None):
        """Request to register this user with the identity service running
           at 'identity_url', using the supplied 'password'. This will
           return a QR code that you must use immediately to add this
           user on the identity service to a QR code generator"""

        if self._username is None:
            return None

        service = self.identity_service()

        args = {"username": self._username,
                "password": password}

        result = service.call_function(function="register", args=args)

        try:
            provisioning_uri = result["provisioning_uri"]
        except:
            from Acquire.Client import UserError
            raise UserError(
                "Cannot register the user '%s' on "
                "the identity service at '%s'!" %
                (self._username, identity_url))

        # return a QR code for the provisioning URI
        from Acquire.Client import create_qrcode as _create_qrcode
        return (provisioning_uri, _create_qrcode(provisioning_uri))

    def request_login(self, login_message=None, _is_local=False):
        """Request to authenticate as this user. This returns a login URL that
           you must connect to to supply your login credentials

           If 'login_message' is supplied, then this is passed to
           the identity service so that it can be displayed
           when the user accesses the login page. This helps
           the user validate that they have accessed the correct
           login page. Note that if the message is None,
           then a random message will be generated.
        """
        self._check_for_error()

        from Acquire.Client import LoginError

        if not self.is_empty():
            raise LoginError("You cannot try to log in twice using the same "
                             "User object. Create another object if you want "
                             "to try to log in again.")

        # first, create a private key that will be used
        # to sign all requests and identify this login
        from Acquire.Client import PrivateKey as _PrivateKey
        session_key = _PrivateKey()
        signing_key = _PrivateKey()

        args = {"username": self._username,
                "public_key": session_key.public_key().to_data(),
                "public_certificate": signing_key.public_key().to_data(),
                "ipaddr": None}

        # get information from the local machine to help
        # the user validate that the login details are correct
        if _has_socket:
            if _is_local:
                args["ipaddr"] = "local"
                args["hostname"] = "local"
            else:
                hostname = _socket.gethostname()
                ipaddr = _socket.gethostbyname(hostname)
                args["ipaddr"] = ipaddr
                args["hostname"] = hostname

        if (login_message is None) and not (_is_local):
            try:
                login_message = \
                    "User '%s' in process '%s' wants to log in..." % \
                    (_os.getlogin(), _os.getpid())
            except:
                # os.getlogin() can fail - message is not needed
                pass

        args["message"] = login_message

        identity_service = self.identity_service()

        args = {"username": self._username,
                "public_key": session_key.public_key().to_data(),
                "public_certificate": signing_key.public_key().to_data(),
                "ipaddr": None,
                "message": login_message}

        result = identity_service.call_function(
                        function="request_login", args=args)

        # look for status = 0
        try:
            status = int(result["status"])
        except:
            status = -1

        try:
            message = result["message"]
        except:
            message = str(result)

        if status != 0:
            error = "Failed to login. Error = %d. Message = %s" % \
                                (status, message)
            self._set_error_state(error)
            raise LoginError(error)

        try:
            login_url = result["login_url"]
        except:
            login_url = None

        if login_url is None:
            error = "Failed to login. Could not extract the login URL! " \
                    "Result is %s" % (str(result))
            self._set_error_state(error)
            raise LoginError(error)

        try:
            session_uid = result["session_uid"]
        except:
            session_uid = None

        if session_uid is None:
            error = "Failed to login. Could not extract the login " \
                    "session UID! Result is %s" % (str(result))

            self._set_error_state(error)
            raise LoginError(error)

        # now save all of the needed data
        self._login_url = result["login_url"]
        self._session_key = session_key
        self._signing_key = signing_key
        self._session_uid = session_uid
        self._status = _LoginStatus.LOGGING_IN
        self._user_uid = result["user_uid"]

        qrcode = None

        from Acquire.Client import has_qrcode as _has_qrcode

        if _has_qrcode() and not _is_local:
            try:
                from Acquire.Client import create_qrcode as _create_qrcode
                self._login_qrcode = _create_qrcode(self._login_url)
                qrcode = self._login_qrcode
            except:
                pass

        return (self._login_url, qrcode)

    def _poll_session_status(self):
        """Function used to query the identity service for this session
           to poll for the session status"""

        service = self.identity_service()

        args = {"username": self._username,
                "session_uid": self._session_uid}

        result = service.call_function(function="get_status", args=args)

        # look for status = 0
        try:
            status = int(result["status"])
        except:
            status = -1

        try:
            message = result["message"]
        except:
            message = str(result)

        if status != 0:
            error = "Failed to query identity service. Error = %d. " \
                    "Message = %s" % (status, message)
            self._set_error_state(error)
            from Acquire.Client import LoginError
            raise LoginError(error)

        # now update the status...
        status = result["session_status"]
        self._set_status(status)

    def wait_for_login(self, timeout=None, polling_delta=5):
        """Block until the user has logged in. If 'timeout' is set
           then we will wait for a maximum of that number of seconds

           This will check whether we have logged in by polling
           the identity service every 'polling_delta' seconds.
        """

        self._check_for_error()

        if not self.is_logging_in():
            return self.is_logged_in()

        polling_delta = int(polling_delta)
        if polling_delta > 60:
            polling_delta = 60
        elif polling_delta < 1:
            polling_delta = 1

        if timeout is None:
            # block forever....
            while True:
                self._poll_session_status()

                if self.is_logged_in():
                    return True

                elif not self.is_logging_in():
                    return False

                _time.sleep(polling_delta)
        else:
            # only block until the timeout has been reached
            timeout = int(timeout)
            if timeout < 1:
                timeout = 1

            start_time = _datetime.now()

            while (_datetime.now() - start_time).seconds < timeout:
                self._poll_session_status()

                if self.is_logged_in():
                    return True

                elif not self.is_logging_in():
                    return False

                _time.sleep(polling_delta)

            return False



"""
Acquire: (C) Christopher Woods 2018

This module provides everything needed to implement an Acquire Accounting
Service

"""

from ._access_service import *
from ._errors import *
from ._checksum import *
from ._request import *
from ._runrequest import *
from ._jobsheet import *

try:
    if __IPYTHON__:
        def _set_printer(C):
            """Function to tell ipython to use __str__ if available"""
            get_ipython().display_formatter.formatters['text/plain'].for_type(
                C,
                lambda obj, p, cycle: p.text(str(obj) if not cycle else '...')
                )

        import sys as _sys
        import inspect as _inspect

        _clsmembers = _inspect.getmembers(_sys.modules[__name__],
                                          _inspect.isclass)

        for _clsmember in _clsmembers:
            _set_printer(_clsmember[1])
except:
    pass




__all__ = ["FileMeta"]


class FileMeta:
    """This is a lightweight class that holds the metadata about
       a particular version of a file
    """
    def __init__(self, filename=None, uid=None, filesize=None,
                 checksum=None, uploaded_by=None, uploaded_when=None,
                 compression=None, aclrules=None):
        """Construct, specifying the filename, and then optionally
           other useful data
        """
        self._filename = filename
        self._uid = uid
        self._filesize = filesize
        self._checksum = checksum
        self._user_guid = uploaded_by
        self._datetime = uploaded_when
        self._compression = compression
        self._acl = None
        self._aclrules = None

        if aclrules is not None:
            from Acquire.Storage import ACLRules as _ACLRules
            if not isinstance(aclrules, _ACLRules):
                raise TypeError("The aclrules must be type ACLRules")

            self._aclrules = aclrules

    def __str__(self):
        """Return a string representation"""
        if self.is_null():
            return "FileMeta::null"
        else:
            return "FileMeta(%s)" % self._filename

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        """Comparison equals"""
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def _set_denied(self):
        """Call this function to remove all information that should
           not be visible to someone who has denied access to the file
        """
        self._uid = None
        self._filesize = None
        self._checksum = None
        self._user_guid = None
        self._datetime = None
        self._compression = None
        self._aclrules = None
        from Acquire.Storage import ACLRule as _ACLRule
        self._acl = _ACLRule.denied()

    def is_null(self):
        """Return whether or not this is null"""
        return self._filename is None

    def has_metadata(self):
        """Return whether or not this file includes all of the
           metadata. If not, then only the filename is available
        """
        return self._uid is not None

    def _set_drive(self, drive):
        """Internal function called by "Drive" to store the Drive
           associated with a FileMeta
        """
        from Acquire.Client import Drive as _Drive
        if not isinstance(drive, _Drive):
            raise TypeError("You can only set a Drive object")

        self._drive = drive

    def filename(self):
        """Return the name of the file"""
        return self._filename

    def uid(self):
        """Return the UID of the file in the system"""
        if self.is_null():
            return None
        else:
            return self._uid

    def filesize(self):
        """If known, return the size of the file"""
        if self.is_null():
            return None
        else:
            return self._filesize

    def checksum(self):
        """If known, return a checksum of the file"""
        if self.is_null():
            return None
        else:
            return self._checksum

    def is_compressed(self):
        """If known, return whether or not this file is stored and
           transmitted in a compressed state
        """
        if self.is_null():
            return False
        else:
            return self._compression is not None

    def compression_type(self):
        """Return the compression type for this file, if it is
           stored and transmitted in a compressed state
        """
        if self.is_null():
            return None
        else:
            return self._compression

    def uploaded_by(self):
        """If known, return the GUID of the user who uploaded
           this version of the file
        """
        if self.is_null():
            return None
        else:
            return self._user_guid

    def uploaded_when(self):
        """If known, return the datetime when this version of
           the file was uploaded
        """
        if self.is_null():
            return None
        else:
            return self._datetime

    def aclrules(self):
        """If known, return the ACL rules that were used to generate the ACL
           for this file. Note that you can only see the ACL rules if
           you are an owner of the file
        """
        try:
            return self._aclrules
        except:
            return None

    def resolve_acl(self, **kwargs):
        """Resolve the ACL for this file based on the passed arguments
           (same as for ACLRules.resolve()). This returns the resolved
           ACL, which is set as self.acl()
        """
        if "resolved_acl" in kwargs:
            acl = kwargs["resolved_acl"]
            from Acquire.Storage import ACLRule as _ACLRule
            if not isinstance(acl, _ACLRule):
                raise TypeError(
                    "The resolved ACL must be type ACLRule")

            self._acl = acl.resolve(must_resolve=True, **kwargs)

            if not self._acl.is_owner():
                # only owners can see the ACLs
                self._aclrules = None

            if self._acl.denied_all():
                # not allowed to know anythign about the file
                self._set_denied()

            return self._acl

        if self._aclrules is None:
            raise PermissionError(
                "You do not have permission to resolve the ACLs for this file")

        self._acl = self._aclrules.resolve(**kwargs)

        if not self._acl.is_owner():
            # only owners can see the ACLs
            self._aclrules = None

        if self._acl.denied_all():
            self._set_denied()

        return self._acl

    def acl(self):
        """If known, return the ACL for this version of the file for the
           user that requested this FileMeta (e.g. the user who listed
           the drive containing this file)
        """
        try:
            return self._acl
        except:
            return None

    def assert_correct_data(self, filedata=None, filename=None):
        """Assert that the passed data is correct (right size and
           checksum)
        """
        if filedata is not None:
            from Acquire.Access import get_size_and_checksum \
                as _get_size_and_checksum
            (filesize, checksum) = _get_size_and_checksum(filedata)
        else:
            from Acquire.Access import get_filesize_and_checksum \
                as _get_filesize_and_checksum
            (filesize, checksum) = _get_filesize_and_checksum(filename)

        if (filesize != self._filesize) or (checksum != self._checksum):
            from Acquire.Storage import FileValidationError
            raise FileValidationError(
                "Possible data corruption. Mismatch in file size or "
                "checksum for file '%s'. "
                "%s versus %s, and %s versus %s" %
                (self._filename, filesize, self._filesize,
                 checksum, self._checksum))

    def to_data(self):
        """Return a json-serialisable dictionary of this object"""
        data = {}

        if self.is_null():
            return data

        data["filename"] = str(self._filename)

        if self._uid is not None:
            data["uid"] = str(self._uid)

        if self._filesize is not None:
            data["filesize"] = self._filesize

        if self._checksum is not None:
            data["checksum"] = self._checksum

        if self._user_guid is not None:
            data["user_guid"] = self._user_guid

        if self._compression is not None:
            data["compression"] = self._compression

        try:
            acl = self._acl
        except:
            acl = None

        if acl is not None:
            data["acl"] = acl.to_data()

        try:
            aclrules = self._aclrules
        except:
            aclrules = None

        if aclrules is not None:
            data["aclrules"] = aclrules.to_data()

        if self._datetime is not None:
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            data["datetime"] = _datetime_to_string(self._datetime)

        return data

    @staticmethod
    def from_data(data):
        """Return a new FileMeta constructed from the passed json-deserialised
           dictionary
        """
        f = FileMeta()

        if data is not None and len(data) > 0:
            f._filename = data["filename"]

            if "uid" in data:
                f._uid = data["uid"]

            if "filesize" in data:
                f._filesize = data["filesize"]

            if "checksum" in data:
                f._checksum = data["checksum"]

            if "user_guid" in data:
                f._user_guid = data["user_guid"]

            if "datetime" in data:
                from Acquire.ObjectStore import string_to_datetime \
                    as _string_to_datetime
                f._datetime = _string_to_datetime(data["datetime"])

            if "compression" in data:
                f._compression = data["compression"]

            if "acl" in data:
                from Acquire.Client import ACLRule as _ACLRule
                f._acl = _ACLRule.from_data(data["acl"])

            if "aclrules" in data:
                from Acquire.Storage import ACLRules as _ACLRules
                f._aclrules = _ACLRules.from_data(data["aclrules"])

        return f




__all__ = ["Refund"]


class Refund:
    """This class holds the refund for a transaction. This is sent
       by the credited account or other authorised party to refund a
       transaction. Refunds must succeed, so potentially they can
       allow accounts to breach overdraft limits etc.
    """
    def __init__(self, credit_note=None, authorisation=None):
        """Create a refund for the transaction that resulted in the passed
           credit note. Specify the authorisation of the refund. Note
           that transactions can only be refunded in full. If you want
           to give a partial refund then this is better handled as
           a new transaction from the credited to debited accounts.

           Note also that you can only refund completed transactions.
           If you need to refund a provisional transaction then receipt
           it first. (Obvs it is easier to just receipt with 0 rather
           than receipt and then refund...)
        """
        if credit_note is None:
            self._credit_note = None
            self._authorisation = None
            return

        from Acquire.Accounting import CreditNote as _CreditNote
        from Acquire.Identity import Authorisation as _Authorisation

        if not isinstance(credit_note, _CreditNote):
            raise TypeError("The credit note must be of type CreditNote")

        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The authorisation must be of type Authorisation")

        self._credit_note = credit_note
        self._authorisation = authorisation

    def __str__(self):
        return "Refund(credit_note=%s)" % str(self.credit_note())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._credit_note == other._credit_note and \
                   self._authorisation == other._authorisation
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_null(self):
        """Return whether or not this Refund is null"""
        return self._credit_note is None

    def credit_note(self):
        """Return the credit note that this is refunding"""
        if self.is_null():
            from Acquire.Accounting import CreditNote as _CreditNote
            return _CreditNote()
        else:
            return self._credit_note

    def transaction_uid(self):
        """Return the UID of the  transaction for which this
           is the refund. The transaction UID is the same as the UID
           for the original debit note
        """
        return self.debit_note_uid()

    def debit_note_uid(self):
        """Return the UID of the debit note that this is refunding"""
        if self.is_null():
            return None
        else:
            return self._credit_note.debit_note_uid()

    def debit_account_uid(self):
        """Return the UID of the account from which this refund will
           return value
        """
        if self.is_null():
            return None
        else:
            return self._credit_note.debit_account_uid()

    def credit_account_uid(self):
        """Return the UID of the account from which this refund will be
           drawn
        """
        if self.is_null():
            return None
        else:
            return self._credit_note.credit_account_uid()

    def transaction(self):
        """Return a transaction that corresponds to the real transfer
           of value back from the credit to debit accounts (remembering
           that the original debit account will be the new credit account,
           and the original credit account will be the new debit account).
        """
        from Acquire.Accounting import Transaction as _Transaction

        if self.is_null():
            return _Transaction()
        else:
            return _Transaction(self.value(),
                                "Refund for transaction %s"
                                % self.transaction_uid())

    def value(self):
        """Return the value of the refund"""
        if self.is_null():
            from Acquire.Accounting import create_decimal as _create_decimal
            return _create_decimal(0)
        else:
            return self._credit_note.value()

    def authorisation(self):
        """Return the authorisation for the refund"""
        return self._authorisation

    def to_data(self):
        """Return the data for this object as a dictionary that can be
           serialised to json
        """
        data = {}

        if not self.is_null():
            data["credit_note"] = self._credit_note.to_data()
            data["authorisation"] = self._authorisation.to_data()

        return data

    @staticmethod
    def from_data(data):
        """Return a Refund from the passed json-decoded dictionary"""
        r = Refund()

        if (data and len(data) > 0):
            from Acquire.Accounting import CreditNote as _CreditNote
            from Acquire.Identity import Authorisation as _Authorisation
            r._credit_note = _CreditNote.from_data(data["credit_note"])
            r._authorisation = _Authorisation.from_data(data["authorisation"])

        return r

    @staticmethod
    def create(credit_notes, authorisation):
        """Construct a series of refunds from the passed credit notes,
           each of which is authorised using the passed authorisation.
           This will refund all of the transactions passed in full
        """
        try:
            credit_note = credit_notes[0]
        except:
            return Refund(credit_notes, authorisation)

        if len(credit_notes) == 0:
            return Refund()
        elif len(credit_notes) == 1:
            return Refund(credit_notes[0], authorisation)

        refunds = []

        for credit_note in credit_notes:
            refunds.append(Refund(credit_note, authorisation))

        return refunds



"""
Acquire : (C) Christopher Woods 2019

This module provides the stubs around modules that are not installed
or available on this system, e.g. for license incompatibility reasons

(i.e. while Acquire is Apache licensed, some users may want to use it
 in systems where GPL modules are installed)
"""

try:
    import lazy_import
    lazy_import.logging.disable(lazy_import.logging.DEBUG)
except:
    # lazy_import is not available, e.g. because we want the Apache
    # licensed version of this code - import the non-lazy wrapper
    from ._lazy_import import *

requests = lazy_import.lazy_module("requests")




from Acquire.Service import ServiceError as _ServiceError

__all__ = ["AccountingServiceError", "LedgerError", "TransactionError",
           "AccountError", "UnbalancedLedgerError", "InsufficientFundsError",
           "UnmatchedReceiptError", "PaymentError"]


class AccountingServiceError(_ServiceError):
    pass


class AccountError(Exception):
    pass


class LedgerError(Exception):
    pass


class TransactionError(Exception):
    pass


class UnbalancedLedgerError(Exception):
    pass


class PaymentError(Exception):
    pass


class InsufficientFundsError(PaymentError):
    pass


class UnmatchedReceiptError(Exception):
    pass


class UnmatchedRefundError(Exception):
    pass




from Acquire.Service import ServiceError as _ServiceError

__all__ = ["AccessServiceError", "RunRequestError"]


class AccessServiceError(_ServiceError):
    pass


class RunRequestError(Exception):
    pass




__all__ = ["DriveMeta"]


class DriveMeta:
    """This is a lightweight class that holds the metadata about
       a Drive
    """
    def __init__(self, name=None, uid=None, container=None, aclrules=None):
        """Construct a drive with a specified name and (optionally) UID.
           'container' is the UID of the drive that contains this drive,
           at least for the user who has accessed this drive via a path.
           If 'container' is none, then the user is accessing this
           drive as a top-level drive
        """
        self._name = name
        self._uid = uid

        if isinstance(container, list):
            if len(container) == 1:
                container = [container]
            elif len(container) == 0:
                container = None

        self._container = container

        if aclrules is not None:
            from Acquire.Storage import ACLRules as _ACLRules
            if not isinstance(aclrules, _ACLRules):
                raise PermissionError(
                    "The ACL rules must be type ACLRules")

        self._aclrules = aclrules
        self._acl = None

    def __str__(self):
        """Return a string representation"""
        if self.is_null():
            return "DriveMeta::null"
        else:
            return "DriveMeta(%s)" % self._name

    def __repr__(self):
        return self.__str__()

    def is_null(self):
        """Return whether or not this is null"""
        return self._name is None

    def name(self):
        """Return the name of the drive"""
        return self._name

    def uid(self):
        """If known, return the UID of the drive"""
        return self._uid

    def acl(self):
        """If known, return the ACL rule for this drive"""
        return self._acl

    def aclrules(self):
        """If known, return the ACL rules for this drive"""
        return self._aclrules

    def resolve_acl(self, **kwargs):
        """Resolve the ACL for this file based on the passed arguments
           (same as for ACLRules.resolve()). This returns the resolved
           ACL, which is set as self.acl()
        """
        if self._aclrules is None:
            raise PermissionError(
                "You do not have permission to resolve the ACLs for "
                "this drive")

        self._acl = self._aclrules.resolve(**kwargs)

        if not self._acl.is_owner():
            # only owners can see the ACLs
            self._aclrules = None

        return self._acl

    def is_top_level(self):
        """Return whether or not this drive was accessed as a
           top-level drive
        """
        return self._container is None

    def container_uid(self):
        """Return the UID of the drive that contains this drive, at
           least via the access path followed by the user.
           This returns None if the user accessed this drive as
           a top-level drive
        """
        if self._container is None:
            return None
        else:
            return self._container[-1]

    def container_uids(self):
        """Return the UIDs of the full hierarchy of the drives that
           contain this drive, as this drive was accessed by the user.
           This returns an empty list if the user accessed this drive
           as a top-level drive
        """
        if self._container is None:
            return []
        else:
            return self._container

    def to_data(self):
        """Return a json-serialisable dictionary of this object"""
        data = {}

        if not self.is_null():
            data["name"] = str(self._name)

            if self._uid is not None:
                data["uid"] = str(self._uid)

            if self._container is not None:
                data["container"] = self._container

            if self._acl is not None:
                data["acl"] = self._acl.to_data()

        return data

    @staticmethod
    def from_data(data):
        """Return an object constructed from the passed json-deserialised
           dictionary
        """

        d = DriveMeta()

        if data is not None and len(data) > 0:
            d._name = data["name"]

            if "uid" in data:
                d._uid = data["uid"]

            if "container" in data:
                d._container = data["container"]

            if "acl" in data:
                from Acquire.Client import ACLRule as _ACLRule
                d._acl = _ACLRule.from_data(data["acl"])

        return d




__all__ = ["Authorisation"]


class Authorisation:
    """This class holds the information needed to show that a user
       has authorised an action. This contains a signed token that
       records the time that the authorisation that was signed, together
       with an extra key (or secret) that can be used by the user
       and provider to verify that the authorisation is for the
       correct resource
    """
    def __init__(self, resource=None, user=None, testing_key=None):
        """Create an authorisation for the passed resource
           that is authorised by the passed user (who must be authenticated)

           If testing_key is passed, then this authorisation is being
           tested as part of the unit tests
        """

        if resource is not None:
            resource = str(resource)

        self._signature = None
        self._last_validated_datetime = None

        if resource is not None:
            if user is None and testing_key is None:
                raise ValueError(
                    "You must pass in an authenticated user who will "
                    "provide authorisation for resource '%s'" % resource)

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now

        if user is not None:
            from Acquire.Client import User as _User

            if not isinstance(user, _User):
                raise TypeError("The passed user must be of type User")

            elif not user.is_logged_in():
                raise PermissionError(
                    "The passed user '%s' must be authenticated to enable "
                    "you to generate an authorisation for the account")

            self._user_uid = user.uid()
            self._session_uid = user.session_uid()
            self._identity_url = user.identity_service().canonical_url()
            self._identity_uid = user.identity_service_uid()
            self._auth_datetime = _get_datetime_now()

            message = self._get_message(resource)
            self._signature = user.signing_key().sign(message)

            self._last_validated_datetime = _get_datetime_now()
            self._last_verified_resource = resource
            self._last_verified_key = None

            if user.guid() != self.user_guid():
                # interesting future case when we allow individual users
                # to be identified by multiple identity services...
                raise PermissionError(
                    "We do not yet support a single user being identified "
                    "by multiple identity services: %s versus %s" %
                    (user.guid(), self.user_guid()))

        elif testing_key is not None:
            self._user_uid = "some user uid"
            self._session_uid = "some session uid"
            self._identity_url = "some identity_url"
            self._identity_uid = "some identity uid"
            self._auth_datetime = _get_datetime_now()
            self._is_testing = True

            message = self._get_message(resource)
            self._signature = testing_key.sign(message)

            self._last_validated_datetime = _get_datetime_now()
            self._last_verified_resource = resource
            self._last_verified_key = testing_key.public_key()

    def is_null(self):
        """Return whether or not this authorisation is null"""
        return self._signature is None

    def _get_message(self, resource=None):
        """Internal function that is used to generate the message for
           the resource that is signed. This message
           encodes information about the user and identity service that
           signed the message, as well as the resource. This helps
           prevent tamporing with the data in this authorisation
        """
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string

        if resource is None:
            return "%s|%s|%s|%s" % (
                self._user_uid, self._session_uid,
                self._identity_uid,
                _datetime_to_string(self._auth_datetime))
        else:
            return "%s|%s|%s|%s|%s" % (
                self._user_uid, self._session_uid,
                self._identity_uid, str(resource),
                _datetime_to_string(self._auth_datetime))

    def __str__(self):
        try:
            return "Authorisation(signature=%s)" % self._signature
        except:
            return "Authorisation()"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._signature == other._signature
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def _fix_integer(self, value, max_value):
        max_value = int(max_value)

        if value is None:
            return max_value
        else:
            try:
                value = int(value)
            except:
                return max_value

        if value <= 0 or value > max_value:
            return max_value
        else:
            return value

    def from_user(self, user_uid, service_uid):
        """Return whether or not this authorisation comes from the user
           with passed user_uid registered on the passed service_uid
        """
        return (user_uid == self._user_uid) and \
               (service_uid == self._identity_uid)

    def user_uid(self):
        """Return the UID of the user who created this authorisation"""
        if self.is_null():
            return None
        else:
            return self._user_uid

    def user_guid(self):
        """Return the global UID for this user"""
        return "%s@%s" % (self.identity_uid(), self.user_uid())

    def session_uid(self):
        """Return the login session that authenticated the user"""
        if self.is_null():
            return None
        else:
            return self._session_uid

    def identity_url(self):
        """Return the URL of the identity service that authenticated
           the user
        """
        if self.is_null():
            return None
        else:
            return self._identity_url

    def identity_uid(self):
        """Return the UID of the identity service that authenticated
           the user
        """
        if self.is_null():
            return None
        else:
            return self._identity_uid

    def signature_time(self):
        """Return the time when the authentication was signed"""
        if self.is_null():
            return None
        else:
            return self._auth_datetime

    def last_verification_time(self):
        """Return the last time this authorisation was verified. Note that
           you should re-verify authorisations periodically, to ensure that
           they identity service is still happy that the login session was
           not suspicious
        """
        if self.is_null():
            return None
        else:
            return self._last_validated_datetime

    def signature(self):
        """Return the actual signature"""
        if self.is_null():
            return None
        else:
            return self._signature

    def is_stale(self, stale_time=7200):
        """Return whether or not this authorisation is stale.
           'stale_time' is the number of seconds after which
           the authorisation is considered stale (and thus
           no longer valid)
        """
        stale_time = self._fix_integer(stale_time, 365*24*7200)

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now

        now = _get_datetime_now()

        return ((now - self._auth_datetime).seconds > stale_time)

    def is_verified(self, resource=None, refresh_time=3600,
                    stale_time=7200, testing_key=None):
        """Return whether or not this authorisation has been verified. Note
           that this will cache any verification for 'refresh_time' (in
           seconds)

           'stale_time' gives the time (in seconds) beyond which the
           authorisation will be considered stale (and thus not valid).
           By default this is 7200 seconds (2 hours), meaning that the
           authorisation must be used within 2 hours to be valid.
        """
        refresh_time = self._fix_integer(refresh_time, 24*3600)

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now

        now = _get_datetime_now()

        try:
            if self._last_verified_resource != resource:
                return False
        except:
            pass

        if self._last_validated_datetime is not None:
            if self._last_verified_key != testing_key:
                return False

            if (now - self._last_validated_datetime).seconds < refresh_time:
                # no need to re-validate
                return not self.is_stale(stale_time)

        return False

    def verify(self, resource=None, refresh_time=3600, stale_time=7200,
               force=False, testing_key=None):
        """Verify that this is a valid authorisation provided by the
           user for the passed 'resource'. This will
           cache the verification for 'refresh_time' (in seconds), but
           re-verification can be forced if 'force' is True.

           'stale_time' gives the time (in seconds) beyond which the
           authorisation will be considered stale (and thus not valid).
           By default this is 7200 seconds (2 hours), meaning that the
           authorisation must be used within 2 hours to be valid.

           If 'testing_key' is passed, then this object is being
           tested as part of the unit tests
        """

        if self.is_null():
            raise PermissionError("Cannot verify a null Authorisation")

        if self.is_stale(stale_time):
            raise PermissionError("Cannot verify a stale Authorisation")

        if not force:
            if self.is_verified(resource=resource,
                                refresh_time=refresh_time,
                                stale_time=stale_time,
                                testing_key=testing_key):
                return

        if testing_key is not None:
            if not self._is_testing:
                raise PermissionError(
                    "You cannot pass a test key to a non-testing "
                    "Authorisation")

            message = self._get_message(resource)

            try:
                testing_key.verify(self._signature, message)
            except Exception as e:
                from Acquire.Service import exception_to_string
                raise PermissionError(exception_to_string(e))

            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now

            self._last_validated_datetime = _get_datetime_now()
            self._last_verified_resource = resource
            self._last_verified_key = testing_key
            return

        try:
            # we need to get the public signing key for this session
            from Acquire.Service import get_trusted_service \
                as _get_trusted_service
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now

            try:
                identity_service = _get_trusted_service(
                                                    self._identity_url)
            except:
                raise PermissionError(
                    "Unable to verify the authorisation as we do not trust "
                    "the identity service at %s" % self._identity_url)

            if not identity_service.can_identify_users():
                raise PermissionError(
                    "Cannot verify an Authorisation that does not use a valid "
                    "identity service")

            if identity_service.uid() != self._identity_uid:
                raise PermissionError(
                    "Cannot verify this Authorisation as the actual UID of "
                    "the identity service at '%s' (%s) does not match "
                    "the UID of the service that signed this authorisation "
                    "(%s)" % (self._identity_url, identity_service.uid(),
                              self._identity_uid))

            response = identity_service.whois(
                                    user_uid=self._user_uid,
                                    session_uid=self._session_uid)

            try:
                logout_datetime = _string_to_datetime(
                                        response["logout_datetime"])
            except:
                logout_datetime = None

            if logout_datetime:
                # the user has logged out from this session - ensure that
                # the authorisation was created before the user logged out
                if logout_datetime < self.signature_time():
                    raise PermissionError(
                        "This authorisation was signed after the user logged "
                        "out. This means that the authorisation is not valid. "
                        "Please log in again and create a new authorisation.")

            message = self._get_message(resource)

            from Acquire.Crypto import PublicKey as _PublicKey
            pubcert = _PublicKey.from_data(response["public_cert"])
            pubcert.verify(self._signature, message)

            self._last_validated_datetime = _get_datetime_now()
            self._last_verified_resource = resource
            self._last_verified_key = None
        except PermissionError:
            raise
        except Exception as e:
            if resource:
                raise PermissionError(
                    "Cannot verify the authorisation for resource %s: %s" %
                    (resource, str(e)))
            else:
                raise PermissionError(
                    "Cannot verify the authorisation: %s" %
                    (str(e)))
        except:
            if resource:
                raise PermissionError(
                    "Cannot verify the authorisation for resource %s" %
                    resource)
            else:
                raise PermissionError("Cannot verify the authorisation")

    @staticmethod
    def from_data(data):
        """Return an authorisation created from the json-decoded dictionary"""
        auth = Authorisation()

        if (data and len(data) > 0):
            from Acquire.ObjectStore import string_to_datetime \
                as _string_to_datetime
            from Acquire.ObjectStore import string_to_bytes \
                as _string_to_bytes

            auth._user_uid = data["user_uid"]
            auth._session_uid = data["session_uid"]
            auth._identity_url = data["identity_url"]
            auth._identity_uid = data["identity_uid"]
            auth._auth_datetime = _string_to_datetime(data["auth_datetime"])
            auth._signature = _string_to_bytes(data["signature"])
            auth._last_validated_datetime = None

            if "is_testing" in data:
                auth._is_testing = data["is_testing"]

        return auth

    def to_data(self):
        """Return this object serialised to a json-encoded dictionary"""
        data = {}

        if self.is_null():
            return data

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import bytes_to_string \
            as _bytes_to_string

        data["user_uid"] = str(self._user_uid)
        data["session_uid"] = str(self._session_uid)
        data["identity_url"] = str(self._identity_url)
        data["identity_uid"] = str(self._identity_uid)
        data["auth_datetime"] = _datetime_to_string(self._auth_datetime)
        data["signature"] = _bytes_to_string(self._signature)

        try:
            data["is_testing"] = self._is_testing
        except:
            pass

        return data




__all__ = ["CreditNote"]


class CreditNote:
    """This class holds all of the information about a completed credit. This
       is combined with a debit note of equal value to form a transaction
       record
    """
    def __init__(self, debit_note=None, account=None,
                 receipt=None, refund=None, bucket=None):
        """Create the corresponding credit note for the passed debit_note. This
           will credit value from the note to the passed account. The credit
           will use the same UID as the debit, and the same datetime. This
           will then be paired with the debit note to form a TransactionRecord
           that can be written to the ledger
        """
        self._account_uid = None

        nargs = (receipt is not None) + (refund is not None)

        if nargs > 1:
            raise ValueError("You can create a CreditNote with a receipt "
                             "or a refund - not both!")

        if receipt is not None:
            self._create_from_receipt(debit_note, receipt, account, bucket)

        elif refund is not None:
            self._create_from_refund(debit_note, refund, account, bucket)

        elif (debit_note is not None) and (account is not None):
            self._create_from_debit_note(debit_note, account, bucket)

        else:
            self._debit_account_uid = None
            self._datetime = None
            self._uid = None
            self._debit_note_uid = None

            from Acquire.Accounting import create_decimal as _create_decimal
            self._value = _create_decimal(0)

    def __str__(self):
        if self.is_null():
            return "CreditNote::null"
        else:
            return "CreditNote:%s>>%s" % (self.value(), self.account_uid())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_null(self):
        """Return whether or not this note is null"""
        return self._uid is None

    def needs_receipting(self):
        """Return whether or not this CreditNote needs to be receipted.
           If so, then the funds are only held provisionally, and must
           be receipted by CreditNote.receipt_by() else they will
           be returned to the DebitNote account
        """
        return self.is_provisional()

    def receipt_by(self):
        """Return the datetime by which this credit note must be
           receipted, or else the funds will be returned. This will
           return None if the CreditNote does not need receipting
        """
        if self.is_provisional():
            return self._receipt_by
        else:
            return None

    def account_uid(self):
        """Return the UID of the account to which the value was credited"""
        if self.is_null():
            return None
        else:
            return self._account_uid

    def credit_account_uid(self):
        """Synonym for self.account_uid()"""
        return self.account_uid()

    def debit_account_uid(self):
        """Return the UID of the account from which the value was debited"""
        if self.is_null():
            return None
        else:
            return self._debit_account_uid

    def datetime(self):
        """Return the datetime for this credit note"""
        return self._datetime

    def uid(self):
        """Return the UID of this credit note. This will not match the debit
           note UID - you need to use debit_note_uid() to get the UID of
           the debit note that matches this credit note
        """
        return self._uid

    def debit_note_uid(self):
        """Return the UID of the debit note that matches this credit note.
           While at the moment only a single credit note matches a debit note,
           it may be in the future that we divide a credit over several
           accounts (and thus several credit notes)
        """
        return self._debit_note_uid

    def value(self):
        """Return the value of this note. This may be less than the
           corresponding debit note if only part of the value of the
           debit note is transferred into the account
        """
        return self._value

    def is_provisional(self):
        """Return whether or not this credit note is provisional
           (i.e. the value will only be transferred on completion
           of work and provision of a receipt. Note that the value
           will only be transferred if this CreditNote is receipted
           before CreditNote.receipt_by())
        """
        if self.is_null():
            return False
        else:
            return self._is_provisional

    def _create_from_refund(self, debit_note, refund, account, bucket):
        """Internal function used to create the credit note from
           the passed refund. This will actually transfer value from the
           debit note to the credited account (which was the original
           debited account)
        """
        from Acquire.Accounting import DebitNote as _DebitNote
        from Acquire.Accounting import Refund as _Refund
        from Acquire.Accounting import TransactionRecord as _TransactionRecord
        from Acquire.Accounting import TransactionState as _TransactionState
        from Acquire.Accounting import Account as _Account

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("You can only create a CreditNote "
                            "with a DebitNote")

        if not isinstance(refund, _Refund):
            raise TypeError("You can only refund a Refund object: %s"
                            % str(refund.__class__))

        # get the transaction behind this refund and ensure it is in the
        # refunding state...
        transaction = _TransactionRecord.load_test_and_set(
                        refund.transaction_uid(),
                        _TransactionState.REFUNDING,
                        _TransactionState.REFUNDING, bucket=bucket)

        # ensure that the receipt matches the transaction...
        transaction.assert_matching_refund(refund)

        if account is None:
            account = _Account(transaction.debit_account_uid(), bucket)
        elif account.uid() != refund.debit_account_uid():
            raise ValueError("The accounts do not match when refunding "
                             "the receipt: %s versus %s" %
                             (account.uid(), refund.debit_account_uid()))

        (uid, datetime) = account._credit_refund(debit_note, refund, bucket)

        self._account_uid = account.uid()
        self._debit_account_uid = debit_note.account_uid()
        self._datetime = datetime
        self._uid = uid
        self._debit_note_uid = debit_note.uid()
        self._value = debit_note.value()
        self._is_provisional = debit_note.is_provisional()

        if self._is_provisional:
            self._receipt_by = debit_note.receipt_by()

        # finally(!) move the transaction into the refunded state
        _TransactionRecord.load_test_and_set(
                            refund.transaction_uid(),
                            _TransactionState.REFUNDING,
                            _TransactionState.REFUNDED, bucket=bucket)

    def _create_from_receipt(self, debit_note, receipt, account, bucket):
        """Internal function used to create the credit note from
           the passed receipt. This will actually transfer value from the
           debit note to the credited account
        """
        from Acquire.Accounting import DebitNote as _DebitNote
        from Acquire.Accounting import Refund as _Refund
        from Acquire.Accounting import TransactionRecord as _TransactionRecord
        from Acquire.Accounting import TransactionState as _TransactionState
        from Acquire.Accounting import Account as _Account
        from Acquire.Accounting import Receipt as _Receipt

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("You can only create a CreditNote "
                            "with a DebitNote")

        if not isinstance(receipt, _Receipt):
            raise TypeError("You can only receipt a Receipt object: %s"
                            % str(receipt.__class__))

        # get the transaction behind this receipt and ensure it is in the
        # receipting state...
        transaction = _TransactionRecord.load_test_and_set(
                        receipt.transaction_uid(),
                        _TransactionState.RECEIPTING,
                        _TransactionState.RECEIPTING, bucket=bucket)

        # ensure that the receipt matches the transaction...
        transaction.assert_matching_receipt(receipt)

        if account is None:
            account = _Account(transaction.credit_account_uid(), bucket)
        elif account.uid() != receipt.credit_account_uid():
            raise ValueError("The accounts do not match when crediting "
                             "the receipt: %s versus %s" %
                             (account.uid(), receipt.credit_account_uid()))

        (uid, datetime) = account._credit_receipt(debit_note, receipt, bucket)

        self._account_uid = account.uid()
        self._debit_account_uid = debit_note.account_uid()
        self._datetime = datetime
        self._uid = uid
        self._debit_note_uid = debit_note.uid()
        self._value = debit_note.value()
        self._is_provisional = debit_note.is_provisional()

        if debit_note.is_provisional():
            self._receipt_by = debit_note.receipt_by()

        # finally(!) move the transaction into the receipted state
        _TransactionRecord.load_test_and_set(
                            receipt.transaction_uid(),
                            _TransactionState.RECEIPTING,
                            _TransactionState.RECEIPTED, bucket=bucket)

    def _create_from_debit_note(self, debit_note, account, bucket):
        """Internal function used to create the credit note that matches
           the passed debit note. This will actually transfer value from
           the debit note to the passed account
        """
        from Acquire.Accounting import DebitNote as _DebitNote
        from Acquire.Accounting import Account as _Account

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("You can only create a CreditNote "
                            "with a DebitNote")

        if not isinstance(account, _Account):
            raise TypeError("You can only create a CreditNote with an "
                            "Account")

        (uid, datetime) = account._credit(debit_note, bucket=bucket)

        self._account_uid = account.uid()
        self._debit_account_uid = debit_note.account_uid()
        self._datetime = datetime
        self._uid = uid
        self._debit_note_uid = debit_note.uid()
        self._value = debit_note.value()
        self._is_provisional = debit_note.is_provisional()

        if self._is_provisional:
            self._receipt_by = debit_note.receipt_by()

    @staticmethod
    def from_data(data):
        """Construct and return a new CreditNote from the passed json-decoded
            dictionary
        """
        note = CreditNote()

        if (data and len(data) > 0):
            from Acquire.ObjectStore import string_to_datetime \
                as _string_to_datetime
            from Acquire.Accounting import create_decimal as _create_decimal

            note._account_uid = data["account_uid"]
            note._debit_account_uid = data["debit_account_uid"]
            note._uid = data["uid"]
            note._debit_note_uid = data["debit_note_uid"]
            note._datetime = _string_to_datetime(data["datetime"])
            note._value = _create_decimal(data["value"])
            note._is_provisional = data["is_provisional"]

            if note._is_provisional:
                note._receipt_by = _string_to_datetime(data["receipt_by"])

        return note

    def to_data(self):
        """Return this credit note as a dictionary that can be
           encoded to json
        """
        data = {}

        if not self.is_null():
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string

            data["account_uid"] = self._account_uid
            data["debit_account_uid"] = self._debit_account_uid
            data["uid"] = self._uid
            data["debit_note_uid"] = self._debit_note_uid
            data["datetime"] = _datetime_to_string(self._datetime)
            data["value"] = str(self._value)
            data["is_provisional"] = self._is_provisional

            if self._is_provisional:
                data["receipt_by"] = _datetime_to_string(self._receipt_by)

        return data




import uuid as _uuid
from copy import copy as _copy

__all__ = ["Ledger"]


class Ledger:
    """This is a static class which manages the global ledger for the
       entire accounting service
    """
    @staticmethod
    def get_key(uid):
        """Return the object store key for the transactionrecord with
           UID=uid
        """
        return "accounting/transactions/%s" % (str(uid))

    @staticmethod
    def load_transaction(uid, bucket=None):
        """Load the transactionrecord with UID=uid from the ledger"""
        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.Accounting import TransactionRecord as _TransactionRecord
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        data = _ObjectStore.get_object_from_json(bucket, Ledger.get_key(uid))

        if data is None:
            from Acquire.Accounting import LedgerError
            raise LedgerError("There is no transaction recorded in the "
                              "ledger with UID=%s (at key %s)" %
                              (uid, Ledger.get_key(uid)))

        return _TransactionRecord.from_data(data)

    @staticmethod
    def save_transaction(record, bucket=None):
        """Save the passed transactionrecord to the object store"""
        from Acquire.Accounting import TransactionRecord as _TransactionRecord

        if not isinstance(record, _TransactionRecord):
            raise TypeError("You can only write TransactionRecord objects "
                            "to the ledger!")

        if not record.is_null():
            if bucket is None:
                from Acquire.Service import get_service_account_bucket \
                    as _get_service_account_bucket
                bucket = _get_service_account_bucket()

            from Acquire.ObjectStore import ObjectStore as _ObjectStore

            _ObjectStore.set_object_from_json(bucket,
                                              Ledger.get_key(record.uid()),
                                              record.to_data())

    @staticmethod
    def refund(refund, bucket=None):
        """Create and record a new transaction from the passed refund. This
           applies the refund, thereby transferring value from the credit
           account to the debit account of the corresponding transaction.
           Note that you can only refund a transaction once!
           This returns the (already recorded) TransactionRecord for the
           refund
        """
        from Acquire.Accounting import Refund as _Refund
        from Acquire.Accounting import Account as _Account
        from Acquire.Accounting import DebitNote as _DebitNote
        from Acquire.Accounting import CreditNote as _CreditNote
        from Acquire.Accounting import PairedNote as _PairedNote
        from Acquire.Accounting import TransactionRecord as _TransactionRecord

        if not isinstance(refund, _Refund):
            raise TypeError("The Refund must be of type Refund")

        if refund.is_null():
            return _TransactionRecord()

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        # return value from the credit to debit accounts
        debit_account = _Account(uid=refund.debit_account_uid(),
                                 bucket=bucket)
        credit_account = _Account(uid=refund.credit_account_uid(),
                                  bucket=bucket)

        # remember that a refund debits from the original credit account...
        # (and can only refund completed (DIRECT) transactions)
        debit_note = _DebitNote(refund=refund, account=credit_account,
                                bucket=bucket)

        # now create the credit note to return the value into the debit account
        try:
            credit_note = _CreditNote(debit_note=debit_note,
                                      refund=refund,
                                      account=debit_account,
                                      bucket=bucket)
        except Exception as e:
            # delete the debit note
            try:
                debit_account._delete_note(debit_note, bucket=bucket)
            except:
                pass

            # reset the transaction to its original state
            try:
                _TransactionRecord.load_test_and_set(
                        refund.transaction_uid(),
                        _TransactionState.REFUNDING,
                        _TransactionState.DIRECT,
                        bucket=bucket)
            except:
                pass

            raise e

        try:
            paired_notes = _PairedNote.create(debit_note, credit_note)
        except Exception as e:
            # delete all records...!
            try:
                debit_account._delete_note(debit_note, bucket=bucket)
            except:
                pass

            try:
                credit_account._delete_note(credit_note, bucket=bucket)
            except:
                pass

            # reset the transaction to the pending state
            try:
                _TransactionRecord.load_test_and_set(
                        refund.transaction_uid(),
                        _TransactionState.REFUNDING,
                        _TransactionState.DIRECT,
                        bucket=bucket)
            except:
                pass

            raise e

        # now record the two entries to the ledger. The below function
        # is guaranteed not to raise an exception
        return Ledger._record_to_ledger(paired_notes, refund=refund,
                                        bucket=bucket)

    @staticmethod
    def receipt(receipt, bucket=None):
        """Create and record a new transaction from the passed receipt. This
           applies the receipt, thereby actually transferring value from the
           debit account to the credit account of the corresponding
           transaction. Note that you can only receipt a transaction once!
           This returns the (already recorded) TransactionRecord for the
           receipt
        """
        from Acquire.Accounting import Receipt as _Receipt
        from Acquire.Accounting import Account as _Account
        from Acquire.Accounting import DebitNote as _DebitNote
        from Acquire.Accounting import CreditNote as _CreditNote
        from Acquire.Accounting import TransactionRecord as _TransactionRecord
        from Acquire.Accounting import PairedNote as _PairedNote

        if not isinstance(receipt, _Receipt):
            raise TypeError("The Receipt must be of type Receipt")

        if receipt.is_null():
            return _TransactionRecord()

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        # extract value into the debit note
        debit_account = _Account(uid=receipt.debit_account_uid(),
                                 bucket=bucket)
        credit_account = _Account(uid=receipt.credit_account_uid(),
                                  bucket=bucket)

        debit_note = _DebitNote(receipt=receipt, account=debit_account,
                                bucket=bucket)

        # now create the credit note to put the value into the credit account
        try:
            credit_note = _CreditNote(debit_note=debit_note,
                                      receipt=receipt,
                                      account=credit_account,
                                      bucket=bucket)
        except Exception as e:
            # delete the debit note
            try:
                debit_account._delete_note(debit_note, bucket=bucket)
            except:
                pass

            # reset the transaction to the pending state
            try:
                _TransactionRecord.load_test_and_set(
                        receipt.transaction_uid(),
                        _TransactionState.RECEIPTING,
                        _TransactionState.PENDING,
                        bucket=bucket)
            except:
                pass

            raise e

        try:
            paired_notes = _PairedNote.create(debit_note, credit_note)
        except Exception as e:
            # delete all records...!
            try:
                debit_account._delete_note(debit_note, bucket=bucket)
            except:
                pass

            try:
                credit_account._delete_note(credit_note, bucket=bucket)
            except:
                pass

            # reset the transaction to the pending state
            try:
                _TransactionRecord.load_test_and_set(
                        receipt.transaction_uid(),
                        _TransactionState.RECEIPTING,
                        _TransactionState.PENDING,
                        bucket=bucket)
            except:
                pass

            raise e

        # now record the two entries to the ledger. The below function
        # is guaranteed not to raise an exception
        return Ledger._record_to_ledger(paired_notes, receipt=receipt,
                                        bucket=bucket)

    @staticmethod
    def perform(transactions, debit_account, credit_account, authorisation,
                is_provisional=False, receipt_by=None, bucket=None):
        """Perform the passed transaction(s) between 'debit_account' and
           'credit_account', recording the 'authorisation' for this
           transaction. If 'is_provisional' then record this as a provisional
           transaction (liability for the debit_account, future unspendable
           income for the 'credit_account'). Payment won't actually be taken
           until the transaction is 'receipted' (which may be for less than
           (but not more than) then provisional value, and which must take
           place before 'receipt_by' (which will default to one week in
           the future if not supplied - the actual time is encoded
           in the returned TransactionRecords). Returns the (already
           recorded) TransactionRecord.

           Note that if several transactions are passed, then they must all
           succeed. If one of them fails then they are immediately refunded.
        """
        from Acquire.Accounting import Account as _Account
        from Acquire.Identity import Authorisation as _Authorisation
        from Acquire.Accounting import DebitNote as _DebitNote
        from Acquire.Accounting import CreditNote as _CreditNote
        from Acquire.Accounting import Transaction as _Transaction
        from Acquire.Accounting import PairedNote as _PairedNote

        if not isinstance(debit_account, _Account):
            raise TypeError("The Debit Account must be of type Account")

        if not isinstance(credit_account, _Account):
            raise TypeError("The Credit Account must be of type Account")

        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The Authorisation must be of type Authorisation")

        if is_provisional:
            is_provisional = True
        else:
            is_provisional = False

        try:
            transactions[0]
        except:
            transactions = [transactions]

        # remove any zero transactions, as they are not worth recording
        t = []
        for transaction in transactions:
            if not isinstance(transaction, _Transaction):
                raise TypeError("The Transaction must be of type Transaction")

            if transaction.value() >= 0:
                t.append(transaction)

        transactions = t

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        # first, try to debit all of the transactions. If any fail (e.g.
        # because there is insufficient balance) then they are all
        # immediately refunded
        debit_notes = []
        try:
            for transaction in transactions:
                debit_notes.append(_DebitNote(transaction, debit_account,
                                              authorisation, is_provisional,
                                              receipt_by, bucket=bucket))

                # ensure the receipt_by date for all notes is the same
                if is_provisional and (receipt_by is None):
                    receipt_by = debit_notes[0].receipt_by()

        except Exception as e:
            # refund all of the completed debits
            credit_notes = []
            debit_error = str(e)
            try:
                for debit_note in debit_notes:
                    debit_account._delete_note(debit_note, bucket=bucket)
            except Exception as e:
                from Acquire.Accounting import UnbalancedLedgerError
                raise UnbalancedLedgerError(
                    "We have an unbalanced ledger as it was not "
                    "possible to refund a multi-part refused credit (%s): "
                    "Credit refusal error = %s. Refund error = %s" %
                    (str(debit_note), str(debit_error), str(e)))

            # raise the original error to show that, e.g. there was
            # insufficient balance
            raise e

        # now create the credit note(s) for this transaction. This will credit
        # the account, thereby transferring value from the debit_note(s) to
        # that account. If this fails then the debit_note(s) needs to
        # be refunded
        credit_notes = {}
        has_error = False
        credit_error = Exception()
        for debit_note in debit_notes:
            try:
                credit_note = _CreditNote(debit_note, credit_account,
                                          bucket=bucket)
                credit_notes[debit_note.uid()] = credit_note
            except Exception as e:
                has_error = True
                credit_error = e
                break

        if has_error:
            # something went wrong crediting the account... We need to refund
            # the transaction - first retract the credit notes...
            try:
                for credit_note in credit_notes.values():
                    credit_account._delete_note(credit_note, bucket=bucket)
            except Exception as e:
                from Acquire.Accounting import UnbalancedLedgerError
                raise UnbalancedLedgerError(
                    "We have an unbalanced ledger as it was not "
                    "possible to credit a multi-part debit (%s): Credit "
                    "refusal error = %s. Refund error = %s" %
                    (debit_notes, str(credit_error), str(e)))

            # now refund all of the debit notes
            try:
                for debit_note in debit_notes:
                    debit_account._delete_note(debit_note, bucket=bucket)
            except Exception as e:
                from Acquire.Accounting import UnbalancedLedgerError
                raise UnbalancedLedgerError(
                    "We have an unbalanced ledger as it was not "
                    "possible to credit a multi-part debit (%s): Credit "
                    "refusal error = %s. Refund error = %s" %
                    (debit_notes, str(credit_error), str(e)))

            raise credit_error

        try:
            paired_notes = _PairedNote.create(debit_notes, credit_notes)
        except Exception as e:
            # delete all of the notes...
            for debit_note in debit_notes:
                try:
                    debit_account._delete_note(debit_note, bucket=bucket)
                except:
                    pass

            for credit_note in credit_notes:
                try:
                    credit_account._delete_note(credit_note, bucket=bucket)
                except:
                    pass

            raise e

        # now write the paired entries to the ledger. The below function
        # is guaranteed not to raise an exception
        return Ledger._record_to_ledger(paired_notes, is_provisional,
                                        bucket=bucket)

    @staticmethod
    def _record_to_ledger(paired_notes, is_provisional=False,
                          receipt=None, refund=None, bucket=None):
        """Internal function used to generate and record transaction records
           from the passed paired debit- and credit-note(s). This will write
           the transaction record(s) to the object store, and will also return
           the record(s).
        """
        from Acquire.Accounting import Receipt as _Receipt
        from Acquire.Accounting import Refund as _Refund
        from Acquire.Accounting import TransactionRecord as _TransactionRecord
        from Acquire.Accounting import TransactionState as _TransactionState

        if receipt is not None:
            if not isinstance(receipt, _Receipt):
                raise TypeError("Receipts must be of type 'Receipt'")

        if refund is not None:
            if not isinstance(refund, _Refund):
                raise TypeError("Refunds must be of type 'Refund'")

        try:
            records = []

            if bucket is None:
                from Acquire.Service import get_service_account_bucket \
                    as _get_service_account_bucket
                bucket = _get_service_account_bucket()

            for paired_note in paired_notes:
                record = _TransactionRecord()
                record._debit_note = paired_note.debit_note()
                record._credit_note = paired_note.credit_note()

                if is_provisional:
                    record._transaction_state = _TransactionState.PROVISIONAL
                else:
                    record._transaction_state = _TransactionState.DIRECT

                if receipt is not None:
                    record._receipt = receipt

                if refund is not None:
                    record._refund = refund

                Ledger.save_transaction(record, bucket)

                records.append(record)

            return records

        except:
            # an error occuring here will break the system, which will
            # require manual cleaning. Mark this as broken!
            try:
                Ledger._set_truly_broken(paired_notes, bucket)
            except:
                pass

            raise SystemError("The ledger is in a very broken state!")

    @staticmethod
    def _set_truly_broken(paired_notes, bucket):
        """Internal function called when an irrecoverable error state
           is detected. This records the notes that caused the error and
           places the affected accounts into an error state
        """
        raise NotImplementedError("_set_truly_broken needs to be implemented!")
